"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createBaseMemoryFsSync = exports.createBaseMemoryFs = exports.createMemoryFs = void 0;
const utils_1 = require("@file-services/utils");
const path_1 = __importDefault(require("@file-services/path"));
const types_1 = require("@file-services/types");
const error_codes_js_1 = require("./error-codes.js");
const posixPath = path_1.default.posix;
/**
 * This is the main function to use, returning a sync/async
 * in-memory file system with extended API.
 *
 * @param rootContents optional data to populate / with
 */
function createMemoryFs(rootContents) {
    const baseFs = createBaseMemoryFs();
    const fs = {
        ...(0, utils_1.createFileSystem)(baseFs),
        root: baseFs.root,
    };
    if (rootContents) {
        fs.populateDirectorySync(posixPath.sep, rootContents);
    }
    return fs;
}
exports.createMemoryFs = createMemoryFs;
/**
 * Utility function to create a *base* sync/async
 * in-memory file system (without the extended API).
 */
function createBaseMemoryFs() {
    const syncMemFs = createBaseMemoryFsSync();
    return { ...syncMemFs, ...(0, utils_1.syncToAsyncFs)(syncMemFs) };
}
exports.createBaseMemoryFs = createBaseMemoryFs;
/**
 * Utility function to create a *base* sync-only
 * in-memory file system (without the extended API).
 */
function createBaseMemoryFsSync() {
    const root = createMemDirectory('memory-fs-root');
    const pathListeners = new utils_1.SetMultiMap();
    const globalListeners = new Set();
    realpathSync.native = realpathSync;
    let workingDirectoryPath = posixPath.sep;
    return {
        root,
        ...posixPath,
        resolve: resolvePath,
        watchService: {
            async watchPath(path, listener) {
                const resolvedPath = resolvePath(path);
                if (listener) {
                    pathListeners.add(resolvedPath, listener);
                }
            },
            async unwatchPath(path, listener) {
                const resolvedPath = resolvePath(path);
                if (listener) {
                    pathListeners.delete(resolvedPath, listener);
                }
                else {
                    pathListeners.deleteKey(resolvedPath);
                }
            },
            async unwatchAllPaths() {
                pathListeners.clear();
            },
            addGlobalListener(listener) {
                globalListeners.add(listener);
            },
            removeGlobalListener(listener) {
                globalListeners.delete(listener);
            },
            clearGlobalListeners() {
                globalListeners.clear();
            },
        },
        caseSensitive: true,
        cwd,
        chdir,
        copyFileSync,
        existsSync,
        lstatSync,
        mkdirSync: mkdirSync,
        readdirSync,
        readFileSync: readFileSync,
        realpathSync,
        readlinkSync,
        renameSync,
        rmdirSync,
        rmSync,
        statSync,
        symlinkSync,
        unlinkSync,
        writeFileSync,
        chmodSync: () => undefined,
    };
    function resolvePath(...pathSegments) {
        return posixPath.resolve(workingDirectoryPath, ...pathSegments);
    }
    function cwd() {
        return workingDirectoryPath;
    }
    function chdir(directoryPath) {
        workingDirectoryPath = resolvePath(directoryPath);
    }
    function readFileSync(filePath, _options) {
        const resolvedPath = resolvePath(filePath);
        const fileNode = getNode(resolvedPath);
        if (!fileNode) {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_FILE, 'ENOENT');
        }
        else if (fileNode.type === 'dir') {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_IS_DIRECTORY, 'EISDIR');
        }
        return fileNode.contents;
    }
    function writeFileSync(filePath, fileContent) {
        if (filePath === '') {
            throw createFsError(filePath, error_codes_js_1.FsErrorCodes.NO_FILE_OR_DIRECTORY, 'ENOENT');
        }
        const resolvedPath = resolvePath(filePath);
        const existingNode = getNode(resolvedPath);
        if (existingNode) {
            if (existingNode.type === 'dir') {
                throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_IS_DIRECTORY, 'EISDIR');
            }
            existingNode.entry = { ...existingNode.entry, mtime: new Date() };
            existingNode.contents = fileContent;
            emitWatchEvent({ path: resolvedPath, stats: existingNode.entry });
        }
        else {
            const parentPath = posixPath.dirname(resolvedPath);
            const parentNode = getNode(parentPath);
            if (!parentNode || parentNode.type !== 'dir') {
                throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.CONTAINING_NOT_EXISTS, 'ENOENT');
            }
            const fileName = posixPath.basename(resolvedPath);
            const currentDate = new Date();
            const newFileNode = {
                type: 'file',
                entry: {
                    name: fileName,
                    birthtime: currentDate,
                    mtime: currentDate,
                    isFile: returnsTrue,
                    isDirectory: returnsFalse,
                    isSymbolicLink: returnsFalse,
                },
                contents: fileContent,
            };
            parentNode.contents.set(fileName, newFileNode);
            emitWatchEvent({ path: resolvedPath, stats: newFileNode.entry });
        }
    }
    function unlinkSync(filePath) {
        const resolvedPath = resolvePath(filePath);
        const parentPath = posixPath.dirname(resolvedPath);
        const parentNode = getNode(parentPath);
        if (!parentNode || parentNode.type !== 'dir') {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_FILE, 'ENOENT');
        }
        const fileName = posixPath.basename(resolvedPath);
        const fileNode = parentNode.contents.get(fileName);
        if (!fileNode) {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_FILE, 'ENOENT');
        }
        else if (fileNode.type === 'dir') {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_IS_DIRECTORY, 'EISDIR');
        }
        parentNode.contents.delete(fileName);
        emitWatchEvent({ path: resolvedPath, stats: null });
    }
    function readdirSync(directoryPath, options) {
        const resolvedPath = resolvePath(directoryPath);
        const directoryNode = getNode(resolvedPath);
        if (!directoryNode) {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_DIRECTORY, 'ENOENT');
        }
        else if (directoryNode.type === 'file') {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_IS_FILE, 'ENOTDIR');
        }
        return !!options && typeof options === 'object' && options.withFileTypes
            ? Array.from(directoryNode.contents.values(), ({ entry }) => entry)
            : Array.from(directoryNode.contents.keys());
    }
    function mkdirSync(directoryPath, options) {
        const resolvedPath = resolvePath(directoryPath);
        const parentPath = posixPath.dirname(resolvedPath);
        let parentNode = getNode(parentPath);
        const recursive = options === null || options === void 0 ? void 0 : options.recursive;
        if (!parentNode) {
            if (recursive) {
                mkdirSync(parentPath, options);
                parentNode = getNode(parentPath);
            }
            else {
                throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.CONTAINING_NOT_EXISTS, 'ENOENT');
            }
        }
        else if (parentNode.type !== 'dir') {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_IS_FILE, 'ENOTDIR');
        }
        else if (parentPath === resolvedPath) {
            if (recursive) {
                return;
            }
            else {
                throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_ALREADY_EXISTS, 'EEXIST');
            }
        }
        const directoryName = posixPath.basename(resolvedPath);
        const existingNode = parentNode.contents.get(directoryName);
        if (existingNode) {
            if (recursive && existingNode.type === 'dir') {
                return;
            }
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_ALREADY_EXISTS, 'EEXIST');
        }
        const newDirNode = createMemDirectory(directoryName);
        parentNode.contents.set(directoryName, newDirNode);
        emitWatchEvent({ path: resolvedPath, stats: newDirNode.entry });
    }
    function rmdirSync(directoryPath) {
        const resolvedPath = resolvePath(directoryPath);
        const parentPath = posixPath.dirname(resolvedPath);
        const parentNode = getNode(parentPath);
        if (!parentNode || parentNode.type !== 'dir') {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_DIRECTORY, 'ENOENT');
        }
        const directoryName = posixPath.basename(resolvedPath);
        const directoryNode = parentNode.contents.get(directoryName);
        if (!directoryNode || directoryNode.type !== 'dir') {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_DIRECTORY, 'ENOENT');
        }
        else if (directoryNode.contents.size > 0) {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.DIRECTORY_NOT_EMPTY, 'ENOTEMPTY');
        }
        parentNode.contents.delete(directoryName);
        emitWatchEvent({ path: resolvedPath, stats: null });
    }
    function rmSync(targetPath, { force, recursive } = {}) {
        const resolvedPath = resolvePath(targetPath);
        const parentPath = posixPath.dirname(resolvedPath);
        const parentNode = getNode(parentPath);
        if (!parentNode || parentNode.type !== 'dir') {
            if (force) {
                return;
            }
            else {
                throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_DIRECTORY, 'ENOENT');
            }
        }
        const targetName = posixPath.basename(resolvedPath);
        const targetNode = parentNode.contents.get(targetName);
        if (!targetNode) {
            if (force) {
                return;
            }
            else {
                throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_DIRECTORY, 'ENOENT');
            }
        }
        if (targetNode.type === 'dir' && !recursive) {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_IS_DIRECTORY, 'EISDIR');
        }
        parentNode.contents.delete(targetName);
        emitWatchEvent({ path: resolvedPath, stats: null });
    }
    function existsSync(nodePath) {
        return !!getNode(resolvePath(nodePath));
    }
    function statSync(nodePath, options) {
        var _a;
        const resolvedPath = resolvePath(nodePath);
        const node = getNode(resolvedPath);
        if (!node) {
            const throwIfNoEntry = (_a = options === null || options === void 0 ? void 0 : options.throwIfNoEntry) !== null && _a !== void 0 ? _a : true;
            if (throwIfNoEntry) {
                throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_FILE_OR_DIRECTORY, 'ENOENT');
            }
            else {
                return undefined;
            }
        }
        return node.entry;
    }
    function lstatSync(nodePath, options) {
        var _a;
        const resolvedPath = resolvePath(nodePath);
        const node = getRawNode(resolvedPath);
        if (!node) {
            const throwIfNoEntry = (_a = options === null || options === void 0 ? void 0 : options.throwIfNoEntry) !== null && _a !== void 0 ? _a : true;
            if (throwIfNoEntry) {
                throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_FILE_OR_DIRECTORY, 'ENOENT');
            }
            else {
                return undefined;
            }
        }
        return node.entry;
    }
    function realpathSync(nodePath) {
        const resolvedPath = resolvePath(nodePath);
        let currentPath = '/';
        let node = root;
        for (const depthName of resolvedPath.split(posixPath.sep)) {
            if (!node) {
                break;
            }
            if (depthName === '') {
                continue;
            }
            if (node.type === 'dir') {
                node = node.contents.get(depthName);
                currentPath = posixPath.join(currentPath, depthName);
                while ((node === null || node === void 0 ? void 0 : node.type) === 'symlink') {
                    currentPath = posixPath.resolve(posixPath.dirname(currentPath), node.target);
                    node = getRawNode(currentPath);
                }
            }
            else {
                node = undefined;
            }
        }
        if (!node) {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_FILE_OR_DIRECTORY, 'ENOENT');
        }
        return currentPath;
    }
    function readlinkSync(nodePath) {
        const resolvedPath = resolvePath(nodePath);
        const node = getRawNode(resolvedPath);
        if (!node) {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.NO_FILE_OR_DIRECTORY, 'ENOENT');
        }
        else if (node.type !== 'symlink') {
            throw createFsError(resolvedPath, error_codes_js_1.FsErrorCodes.PATH_IS_INVALID, 'EINVAL');
        }
        return node.target;
    }
    function renameSync(sourcePath, destinationPath) {
        const resolvedSourcePath = resolvePath(sourcePath);
        const resolvedDestinationPath = resolvePath(destinationPath);
        const sourceParentPath = posixPath.dirname(resolvedSourcePath);
        const sourceParentNode = getNode(sourceParentPath);
        if (!sourceParentNode || sourceParentNode.type !== 'dir') {
            throw createFsError(resolvedSourcePath, error_codes_js_1.FsErrorCodes.CONTAINING_NOT_EXISTS, 'ENOENT');
        }
        const sourceName = posixPath.basename(resolvedSourcePath);
        const sourceNode = sourceParentNode.contents.get(sourceName);
        if (!sourceNode) {
            throw createFsError(resolvedSourcePath, error_codes_js_1.FsErrorCodes.NO_FILE_OR_DIRECTORY, 'ENOENT');
        }
        const destinationParentPath = posixPath.dirname(resolvedDestinationPath);
        const destinationParentNode = getNode(destinationParentPath);
        if (!destinationParentNode || destinationParentNode.type !== 'dir') {
            throw createFsError(resolvedDestinationPath, error_codes_js_1.FsErrorCodes.CONTAINING_NOT_EXISTS, 'ENOENT');
        }
        const destinationName = posixPath.basename(resolvedDestinationPath);
        const destinationNode = destinationParentNode.contents.get(destinationName);
        if (destinationNode) {
            if (destinationNode.type === 'dir') {
                if (destinationNode.contents.size > 0) {
                    throw createFsError(resolvedDestinationPath, error_codes_js_1.FsErrorCodes.DIRECTORY_NOT_EMPTY, 'ENOTEMPTY');
                }
            }
            else {
                throw createFsError(resolvedDestinationPath, error_codes_js_1.FsErrorCodes.PATH_ALREADY_EXISTS, 'EEXIST');
            }
        }
        sourceParentNode.contents.delete(sourceName);
        sourceNode.entry = { ...sourceNode.entry, name: destinationName, mtime: new Date() };
        destinationParentNode.contents.set(destinationName, sourceNode);
        emitWatchEvent({ path: resolvedSourcePath, stats: null });
        emitWatchEvent({ path: resolvedDestinationPath, stats: sourceNode.entry });
    }
    function copyFileSync(sourcePath, destinationPath, flags = 0) {
        const resolvedSourcePath = resolvePath(sourcePath);
        const resolvedDestinationPath = resolvePath(destinationPath);
        const sourceFileNode = getNode(resolvedSourcePath);
        if (!sourceFileNode) {
            throw createFsError(resolvedSourcePath, error_codes_js_1.FsErrorCodes.NO_FILE_OR_DIRECTORY, 'ENOENT');
        }
        if (sourceFileNode.type !== 'file') {
            throw createFsError(resolvedSourcePath, error_codes_js_1.FsErrorCodes.PATH_IS_DIRECTORY, 'EISDIR');
        }
        const destParentPath = posixPath.dirname(resolvedDestinationPath);
        const destParentNode = getNode(destParentPath);
        if (!destParentNode || destParentNode.type !== 'dir') {
            throw createFsError(resolvedDestinationPath, error_codes_js_1.FsErrorCodes.CONTAINING_NOT_EXISTS, 'ENOENT');
        }
        const targetName = posixPath.basename(resolvedDestinationPath);
        const destinationFileNode = destParentNode.contents.get(targetName);
        if (destinationFileNode) {
            const shouldOverride = !(flags & types_1.FileSystemConstants.COPYFILE_EXCL);
            if (!shouldOverride) {
                throw createFsError(resolvedDestinationPath, error_codes_js_1.FsErrorCodes.PATH_ALREADY_EXISTS, 'EEXIST');
            }
            if (destinationFileNode.type !== 'file') {
                throw createFsError(resolvedDestinationPath, error_codes_js_1.FsErrorCodes.PATH_IS_DIRECTORY, 'EISDIR');
            }
        }
        const newFileNode = {
            ...sourceFileNode,
            entry: { ...sourceFileNode.entry, name: targetName, mtime: new Date() },
        };
        destParentNode.contents.set(targetName, newFileNode);
        emitWatchEvent({ path: resolvedDestinationPath, stats: newFileNode.entry });
    }
    function symlinkSync(target, linkPath) {
        const resolvedLinkPath = resolvePath(linkPath);
        if (getNode(resolvedLinkPath)) {
            throw createFsError(resolvedLinkPath, error_codes_js_1.FsErrorCodes.PATH_ALREADY_EXISTS, 'EEXIST');
        }
        const parentLinkPath = posixPath.dirname(resolvedLinkPath);
        const parentNode = getNode(parentLinkPath);
        if (!parentNode) {
            throw createFsError(resolvedLinkPath, error_codes_js_1.FsErrorCodes.NO_FILE, 'ENOENT');
        }
        if (parentNode.type === 'file') {
            throw createFsError(resolvedLinkPath, error_codes_js_1.FsErrorCodes.PATH_IS_FILE, 'ENOTDIR');
        }
        const currentDate = new Date(Date.now());
        const fileName = posixPath.basename(resolvedLinkPath);
        const symlinkNode = {
            type: 'symlink',
            entry: {
                name: fileName,
                birthtime: currentDate,
                mtime: currentDate,
                isFile: returnsFalse,
                isDirectory: returnsFalse,
                isSymbolicLink: returnsTrue,
            },
            target,
        };
        parentNode.contents.set(fileName, symlinkNode);
        emitWatchEvent({ path: resolvedLinkPath, stats: symlinkNode.entry });
    }
    function getNode(nodePath) {
        let node = getRawNode(nodePath);
        while ((node === null || node === void 0 ? void 0 : node.type) === 'symlink') {
            nodePath = posixPath.resolve(posixPath.dirname(nodePath), node.target);
            node = getRawNode(nodePath);
        }
        return node;
    }
    function getRawNode(nodePath) {
        let currentPath = '/';
        let node = root;
        for (const depthName of nodePath.split(posixPath.sep)) {
            if (!node) {
                break;
            }
            if (depthName === '') {
                continue;
            }
            while ((node === null || node === void 0 ? void 0 : node.type) === 'symlink') {
                currentPath = posixPath.resolve(posixPath.dirname(currentPath), node.target);
                node = getRawNode(currentPath);
            }
            if ((node === null || node === void 0 ? void 0 : node.type) === 'dir') {
                node = node.contents.get(depthName);
                currentPath = posixPath.join(currentPath, depthName);
            }
            else {
                node = undefined;
            }
        }
        return node;
    }
    function emitWatchEvent(watchEvent) {
        for (const listener of globalListeners) {
            listener(watchEvent);
        }
        const listeners = pathListeners.get(watchEvent.path);
        if (listeners) {
            for (const listener of listeners) {
                listener(watchEvent);
            }
        }
    }
}
exports.createBaseMemoryFsSync = createBaseMemoryFsSync;
function createMemDirectory(name) {
    const currentDate = new Date();
    return {
        type: 'dir',
        contents: new Map(),
        entry: {
            name,
            birthtime: currentDate,
            mtime: currentDate,
            isFile: returnsFalse,
            isDirectory: returnsTrue,
            isSymbolicLink: returnsFalse,
        },
    };
}
const returnsTrue = () => true;
const returnsFalse = () => false;
function createFsError(path, message, code) {
    const error = new Error(`${path} ${message}`);
    error.path = path;
    error.code = code;
    throw error;
}
//# sourceMappingURL=memory-fs.js.map