"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createOverlayFs = void 0;
const utils_1 = require("@file-services/utils");
const getEntryName = (item) => item.name;
function createOverlayFs(lowerFs, upperFs, baseDirectoryPath = lowerFs.cwd()) {
    const { promises: lowerPromises } = lowerFs;
    const { promises: upperPromises } = upperFs;
    const lowerFsRelativeUp = `..${lowerFs.sep}`;
    // ensure base Directory is absolute
    baseDirectoryPath = lowerFs.resolve(baseDirectoryPath);
    function resolvePaths(path) {
        const resolvedLowerPath = lowerFs.resolve(path);
        const relativeToBase = lowerFs.relative(baseDirectoryPath, resolvedLowerPath);
        if (relativeToBase !== '..' &&
            !relativeToBase.startsWith(lowerFsRelativeUp) &&
            !lowerFs.isAbsolute(lowerFsRelativeUp)) {
            return { resolvedLowerPath, resolvedUpperPath: relativeToBase.replace(/\\/g, '/') };
        }
        else {
            return { resolvedLowerPath };
        }
    }
    function realpathSync(path) {
        const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
        if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
            const { stackTraceLimit } = Error;
            try {
                Error.stackTraceLimit = 0;
                return lowerFs.join(baseDirectoryPath, upperFs.realpathSync(resolvedUpperPath));
            }
            catch {
                /**/
            }
            finally {
                Error.stackTraceLimit = stackTraceLimit;
            }
        }
        return lowerFs.realpathSync(resolvedLowerPath);
    }
    realpathSync.native = function realpathSyncNative(path) {
        const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
        if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
            const { stackTraceLimit } = Error;
            try {
                Error.stackTraceLimit = 0;
                return lowerFs.join(baseDirectoryPath, upperFs.realpathSync.native(resolvedUpperPath));
            }
            catch {
                /**/
            }
            finally {
                Error.stackTraceLimit = stackTraceLimit;
            }
        }
        return lowerFs.realpathSync.native(resolvedLowerPath);
    };
    const baseSyncActions = {
        existsSync(path) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                return upperFs.existsSync(resolvedUpperPath) || lowerFs.existsSync(resolvedLowerPath);
            }
            else {
                return lowerFs.existsSync(resolvedLowerPath);
            }
        },
        readFileSync: function readFileSync(path, ...args) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                const { stackTraceLimit } = Error;
                try {
                    Error.stackTraceLimit = 0;
                    return upperFs.readFileSync(resolvedUpperPath, ...args);
                }
                catch {
                    /**/
                }
                finally {
                    Error.stackTraceLimit = stackTraceLimit;
                }
            }
            return lowerFs.readFileSync(resolvedLowerPath, ...args);
        },
        statSync: function (path, ...args) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                try {
                    const stats = upperFs.statSync(resolvedUpperPath, ...args);
                    if (stats) {
                        return stats;
                    }
                }
                catch {
                    /**/
                }
            }
            return lowerFs.statSync(resolvedLowerPath, ...args);
        },
        lstatSync: function lstatSync(path, ...args) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                try {
                    const stats = upperFs.lstatSync(resolvedUpperPath, ...args);
                    if (stats) {
                        return stats;
                    }
                }
                catch {
                    /**/
                }
            }
            return lowerFs.lstatSync(resolvedLowerPath, ...args);
        },
        realpathSync,
        readlinkSync(path) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                const { stackTraceLimit } = Error;
                try {
                    Error.stackTraceLimit = 0;
                    return upperFs.readlinkSync(resolvedUpperPath);
                }
                catch {
                    /**/
                }
                finally {
                    Error.stackTraceLimit = stackTraceLimit;
                }
            }
            return lowerFs.readlinkSync(resolvedLowerPath);
        },
        readdirSync: ((path, options) => {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                const { stackTraceLimit } = Error;
                try {
                    Error.stackTraceLimit = 0;
                    const resInUpper = upperFs.readdirSync(resolvedUpperPath, options);
                    try {
                        const resInLower = lowerFs.readdirSync(resolvedLowerPath, options);
                        if (options !== null && typeof options === 'object' && options.withFileTypes) {
                            const namesInUpper = new Set(resInUpper.map(getEntryName));
                            return [...resInLower.filter((item) => !namesInUpper.has(item.name)), ...resInUpper];
                        }
                        return Array.from(new Set([...resInLower, ...resInUpper]));
                    }
                    catch {
                        return resInUpper;
                    }
                }
                catch {
                    /**/
                }
                finally {
                    Error.stackTraceLimit = stackTraceLimit;
                }
            }
            return lowerFs.readdirSync(resolvedLowerPath, options);
        }),
    };
    const basePromiseActions = {
        async exists(path) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                return (await upperPromises.exists(resolvedUpperPath)) || (await lowerPromises.exists(resolvedLowerPath));
            }
            else {
                return lowerPromises.exists(resolvedLowerPath);
            }
        },
        readFile: async function readFile(path, ...args) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                try {
                    return await upperPromises.readFile(resolvedUpperPath, ...args);
                }
                catch {
                    /**/
                }
            }
            return lowerPromises.readFile(resolvedLowerPath, ...args);
        },
        async stat(path) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                try {
                    return await upperPromises.stat(resolvedUpperPath);
                }
                catch {
                    /**/
                }
            }
            return lowerPromises.stat(resolvedLowerPath);
        },
        async lstat(path) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                try {
                    return await upperPromises.lstat(resolvedUpperPath);
                }
                catch {
                    /**/
                }
            }
            return lowerPromises.lstat(resolvedLowerPath);
        },
        async realpath(path) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                try {
                    return lowerFs.join(baseDirectoryPath, await upperPromises.realpath(resolvedUpperPath));
                }
                catch {
                    /**/
                }
            }
            return lowerPromises.realpath(resolvedLowerPath);
        },
        async readlink(path) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                try {
                    return await upperPromises.readlink(resolvedUpperPath);
                }
                catch {
                    /**/
                }
            }
            return lowerPromises.readlink(resolvedLowerPath);
        },
        readdir: async function readdir(path, options) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined && upperFs.existsSync(resolvedUpperPath)) {
                try {
                    const resInUpper = await upperPromises.readdir(resolvedUpperPath, options);
                    try {
                        const resInLower = await lowerPromises.readdir(resolvedLowerPath, options);
                        if (options !== null && typeof options === 'object' && options.withFileTypes) {
                            const namesInUpper = new Set(resInUpper.map(getEntryName));
                            return [...resInLower.filter((item) => !namesInUpper.has(item.name)), ...resInUpper];
                        }
                        return Array.from(new Set([...resInLower, ...resInUpper]));
                    }
                    catch {
                        /**/
                    }
                    return resInUpper;
                }
                catch {
                    /**/
                }
            }
            return lowerPromises.readdir(resolvedLowerPath, options);
        },
    };
    const baseCallbackActions = {
        exists(path, callback) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                upperFs.exists(resolvedUpperPath, (pathExists) => {
                    if (pathExists) {
                        callback(pathExists);
                    }
                    else {
                        lowerFs.exists(resolvedLowerPath, callback);
                    }
                });
            }
            else {
                lowerFs.exists(resolvedLowerPath, callback);
            }
        },
        readFile(path, options, callback) {
            if (typeof options === 'function') {
                callback = options;
                options = undefined;
            }
            else if (typeof callback !== 'function') {
                throw new Error(`callback is not a function.`);
            }
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                upperFs.readFile(resolvedUpperPath, options, (upperError, upperValue) => {
                    if (upperError) {
                        lowerFs.readFile(resolvedLowerPath, options, callback);
                    }
                    else {
                        callback(upperError, upperValue);
                    }
                });
            }
            else {
                lowerFs.readFile(resolvedLowerPath, options, callback);
            }
        },
        stat(path, callback) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                upperFs.stat(resolvedUpperPath, (e, stats) => {
                    if (e) {
                        lowerFs.stat(resolvedLowerPath, callback);
                    }
                    else {
                        callback(e, stats);
                    }
                });
            }
            else {
                lowerFs.stat(resolvedLowerPath, callback);
            }
        },
        lstat(path, callback) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                upperFs.lstat(resolvedUpperPath, (e, stats) => {
                    if (e) {
                        lowerFs.lstat(resolvedLowerPath, callback);
                    }
                    else {
                        callback(e, stats);
                    }
                });
            }
            else {
                lowerFs.lstat(resolvedLowerPath, callback);
            }
        },
        realpath(path, callback) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                upperFs.realpath(resolvedUpperPath, (e, realPath) => {
                    if (e) {
                        lowerFs.realpath(resolvedLowerPath, callback);
                    }
                    else {
                        callback(e, lowerFs.join(baseDirectoryPath, realPath));
                    }
                });
            }
            else {
                lowerFs.realpath(resolvedLowerPath, callback);
            }
        },
        readlink(path, callback) {
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                upperFs.readlink(resolvedUpperPath, (e, linkPath) => {
                    if (e) {
                        lowerFs.readlink(resolvedLowerPath, callback);
                    }
                    else {
                        callback(e, linkPath);
                    }
                });
            }
            else {
                lowerFs.readlink(resolvedLowerPath, callback);
            }
        },
        readdir(path, options, callback) {
            if (typeof options === 'function') {
                callback = options;
                options = undefined;
            }
            else if (typeof callback !== 'function') {
                throw new Error(`callback is not a function.`);
            }
            const { resolvedLowerPath, resolvedUpperPath } = resolvePaths(path);
            if (resolvedUpperPath !== undefined) {
                upperFs.readdir(resolvedUpperPath, options, (upperError, resInUpper) => {
                    if (upperError) {
                        lowerFs.readdir(resolvedLowerPath, options, callback);
                    }
                    else {
                        lowerFs.readdir(resolvedLowerPath, options, (lowerError, resInLower) => {
                            if (lowerError) {
                                callback(upperError, resInUpper);
                            }
                            else {
                                if (options !== null && typeof options === 'object' && options.withFileTypes) {
                                    const namesInUpper = new Set(resInUpper.map(getEntryName));
                                    const combined = [...resInLower.filter((item) => !namesInUpper.has(item.name)), ...resInUpper];
                                    callback(upperError, combined);
                                }
                                else {
                                    const combined = Array.from(new Set([...resInLower, ...resInUpper]));
                                    callback(upperError, combined);
                                }
                            }
                        });
                    }
                });
            }
            else {
                lowerFs.readdir(resolvedLowerPath, options, callback);
            }
        },
    };
    return (0, utils_1.createFileSystem)({
        ...lowerFs,
        ...baseSyncActions,
        ...baseCallbackActions,
        promises: { ...lowerPromises, ...basePromiseActions },
    });
}
exports.createOverlayFs = createOverlayFs;
//# sourceMappingURL=overlay-fs.js.map