"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createRequestRemapper = exports.createRequestResolver = void 0;
const defaultTarget = 'browser';
const defaultPackageRoots = ['node_modules'];
const defaultExtensions = ['.js', '.json'];
const isRelative = (request) => request === '.' || request === '..' || request.startsWith('./') || request.startsWith('../');
const PACKAGE_JSON = 'package.json';
const statsNoThrowOptions = { throwIfNoEntry: false };
function createRequestResolver(options) {
    const { fs: { statSync, readFileSync, realpathSync, dirname, join, resolve, isAbsolute }, packageRoots = defaultPackageRoots, extensions = defaultExtensions, target = defaultTarget, resolvedPacakgesCache = new Map(), alias = {}, fallback = {}, } = options;
    const loadPackageJsonFromCached = wrapWithCache(loadPackageJsonFrom, resolvedPacakgesCache);
    const remapUsingAlias = createRequestRemapper(alias);
    const remapUsingFallback = createRequestRemapper(fallback);
    return requestResolver;
    function requestResolver(contextPath, originalRequest) {
        var _a, _b;
        const visitedPaths = new Set();
        for (const request of requestsToTry(contextPath, originalRequest, visitedPaths)) {
            if (request === false) {
                return { resolvedFile: request, visitedPaths };
            }
            for (const resolvedFilePath of nodeRequestPaths(contextPath, request)) {
                visitedPaths.add(resolvedFilePath);
                if (!((_a = statSyncSafe(resolvedFilePath)) === null || _a === void 0 ? void 0 : _a.isFile())) {
                    continue;
                }
                if (target === 'browser') {
                    const toPackageJson = findUpPackageJson(dirname(resolvedFilePath));
                    if (toPackageJson) {
                        visitedPaths.add(toPackageJson.filePath);
                        const remappedFilePath = (_b = toPackageJson.browserMappings) === null || _b === void 0 ? void 0 : _b[resolvedFilePath];
                        if (remappedFilePath !== undefined) {
                            if (remappedFilePath !== false) {
                                visitedPaths.add(remappedFilePath);
                            }
                            return {
                                resolvedFile: remappedFilePath,
                                originalFilePath: resolvedFilePath,
                                visitedPaths,
                            };
                        }
                    }
                }
                const realResolvedFilePath = realpathSyncSafe(resolvedFilePath);
                visitedPaths.add(realResolvedFilePath);
                return { resolvedFile: realResolvedFilePath, visitedPaths };
            }
        }
        return { resolvedFile: undefined, visitedPaths };
    }
    function* requestsToTry(contextPath, request, visitedPaths) {
        var _a;
        const requestAlias = remapUsingAlias(request);
        let emittedCandidate = false;
        if (requestAlias !== undefined) {
            emittedCandidate = true;
            yield requestAlias;
        }
        if (!emittedCandidate && target === 'browser') {
            const fromPackageJson = findUpPackageJson(contextPath);
            if (fromPackageJson) {
                visitedPaths.add(fromPackageJson.filePath);
                const remappedRequest = (_a = fromPackageJson.browserMappings) === null || _a === void 0 ? void 0 : _a[request];
                if (remappedRequest !== undefined) {
                    emittedCandidate = true;
                    yield remappedRequest;
                }
            }
        }
        if (!emittedCandidate) {
            yield request;
        }
        const requestFallback = remapUsingFallback(request);
        if (requestFallback !== undefined) {
            yield requestFallback;
        }
    }
    function* nodeRequestPaths(contextPath, request) {
        if (isRelative(request) || isAbsolute(request)) {
            const requestPath = resolve(contextPath, request);
            yield* fileRequestPaths(requestPath);
            yield* directoryRequestPaths(requestPath);
        }
        else {
            yield* packageRequestPaths(contextPath, request);
        }
    }
    /**
     * /path/to/target
     * /path/to/target.js
     * /path/to/target.json
     */
    function* fileRequestPaths(filePath) {
        yield filePath;
        for (const ext of extensions) {
            yield filePath + ext;
        }
    }
    /**
     * /path/to/target (+ext)
     * /path/to/target/index (+ext)
     */
    function* fileOrDirIndexRequestPaths(targetPath) {
        yield* fileRequestPaths(targetPath);
        yield* fileRequestPaths(join(targetPath, 'index'));
    }
    function* directoryRequestPaths(directoryPath) {
        var _a;
        if (!((_a = statSyncSafe(directoryPath)) === null || _a === void 0 ? void 0 : _a.isDirectory())) {
            return;
        }
        const resolvedPackageJson = loadPackageJsonFromCached(directoryPath);
        const mainPath = resolvedPackageJson === null || resolvedPackageJson === void 0 ? void 0 : resolvedPackageJson.mainPath;
        if (mainPath !== undefined) {
            yield* fileOrDirIndexRequestPaths(join(directoryPath, mainPath));
        }
        else {
            yield* fileRequestPaths(join(directoryPath, 'index'));
        }
    }
    function* packageRequestPaths(initialPath, request) {
        var _a;
        for (const packagesPath of packageRootsToPaths(initialPath)) {
            if (!((_a = statSyncSafe(packagesPath)) === null || _a === void 0 ? void 0 : _a.isDirectory())) {
                continue;
            }
            const requestInPackages = join(packagesPath, request);
            yield* fileRequestPaths(requestInPackages);
            yield* directoryRequestPaths(requestInPackages);
        }
    }
    function* packageRootsToPaths(initialPath) {
        for (const packageRoot of packageRoots) {
            for (const directoryPath of pathChainToRoot(initialPath)) {
                yield join(directoryPath, packageRoot);
            }
        }
    }
    function findUpPackageJson(initialPath) {
        for (const directoryPath of pathChainToRoot(initialPath)) {
            const resolvedPackageJson = loadPackageJsonFromCached(directoryPath);
            if (resolvedPackageJson) {
                return resolvedPackageJson;
            }
        }
        return undefined;
    }
    function loadPackageJsonFrom(directoryPath) {
        const packageJsonPath = join(directoryPath, PACKAGE_JSON);
        const packageJson = readJsonFileSyncSafe(packageJsonPath);
        if (typeof packageJson !== 'object' || packageJson === null) {
            return undefined;
        }
        const mainPath = packageJsonTarget(packageJson);
        const { browser } = packageJson;
        let browserMappings = undefined;
        if (target === 'browser' && typeof browser === 'object' && browser !== null) {
            browserMappings = Object.create(null);
            for (const [from, to] of Object.entries(browser)) {
                const resolvedFrom = isRelative(from) ? resolveRelative(join(directoryPath, from)) : from;
                if (resolvedFrom && to !== undefined) {
                    const resolvedTo = resolveRemappedRequest(directoryPath, to);
                    if (resolvedTo !== undefined) {
                        browserMappings[resolvedFrom] = resolvedTo;
                    }
                }
            }
        }
        return {
            filePath: packageJsonPath,
            directoryPath,
            mainPath,
            browserMappings,
        };
    }
    function resolveRemappedRequest(directoryPath, to) {
        if (to === false) {
            return to;
        }
        else if (typeof to === 'string') {
            if (isRelative(to)) {
                return resolveRelative(join(directoryPath, to));
            }
            else {
                return to;
            }
        }
        return undefined;
    }
    function resolveRelative(request) {
        var _a;
        for (const filePath of fileOrDirIndexRequestPaths(request)) {
            if ((_a = statSyncSafe(filePath)) === null || _a === void 0 ? void 0 : _a.isFile()) {
                return realpathSyncSafe(filePath);
            }
        }
        return undefined;
    }
    function statSyncSafe(path) {
        const { stackTraceLimit } = Error;
        try {
            Error.stackTraceLimit = 0;
            return statSync(path, statsNoThrowOptions);
        }
        catch {
            return undefined;
        }
        finally {
            Error.stackTraceLimit = stackTraceLimit;
        }
    }
    function realpathSyncSafe(itemPath) {
        const { stackTraceLimit } = Error;
        try {
            Error.stackTraceLimit = 0;
            return realpathSync(itemPath);
        }
        catch {
            return itemPath;
        }
        finally {
            Error.stackTraceLimit = stackTraceLimit;
        }
    }
    function packageJsonTarget({ main, browser }) {
        if (target === 'browser' && typeof browser === 'string') {
            return browser;
        }
        else {
            return typeof main === 'string' ? main : undefined;
        }
    }
    function* pathChainToRoot(currentPath) {
        let lastPath;
        while (lastPath !== currentPath) {
            yield currentPath;
            lastPath = currentPath;
            currentPath = dirname(currentPath);
        }
    }
    function readJsonFileSyncSafe(filePath) {
        const { stackTraceLimit } = Error;
        try {
            Error.stackTraceLimit = 0;
            return JSON.parse(readFileSync(filePath, 'utf8'));
        }
        catch {
            return undefined;
        }
        finally {
            Error.stackTraceLimit = stackTraceLimit;
        }
    }
}
exports.createRequestResolver = createRequestResolver;
function wrapWithCache(fn, cache = new Map()) {
    return (key) => {
        if (cache.has(key)) {
            return cache.get(key);
        }
        else {
            const result = fn(key);
            cache.set(key, result);
            return result;
        }
    };
}
/**
 * Create a function that accepts a string and returns T | undefined.
 * The remapper supports paths ending with "/*", both in key and value.
 */
function createRequestRemapper(mapping) {
    const parsedTemplateMap = new Map();
    let hasTemplate = false;
    for (const [key, value] of Object.entries(mapping)) {
        let parsedKey = key;
        let parsedValue = value;
        if (key.endsWith('/*')) {
            hasTemplate = true;
            parsedKey = { prefix: key.slice(0, -1) };
            if (typeof value === 'string' && value.endsWith('/*')) {
                parsedValue = { prefix: value.slice(0, -1) };
            }
        }
        parsedTemplateMap.set(parsedKey, parsedValue);
    }
    return hasTemplate
        ? (request) => getFromParsedTemplateMap(parsedTemplateMap, request)
        : (request) => parsedTemplateMap.get(request);
}
exports.createRequestRemapper = createRequestRemapper;
function getFromParsedTemplateMap(map, request) {
    for (const [key, value] of map) {
        const keyType = typeof key;
        if (keyType === 'string') {
            if (request === key) {
                return value;
            }
        }
        else if (keyType === 'object') {
            const { prefix: keyPrefix } = key;
            if (request.startsWith(keyPrefix) && request.length > keyPrefix.length) {
                return typeof value === 'object' ? (value.prefix + request.slice(keyPrefix.length)) : value;
            }
        }
    }
    return undefined;
}
//# sourceMappingURL=request-resolver.js.map