"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createExtendedFileSystemPromiseActions = exports.createAsyncFileSystem = exports.createExtendedSyncActions = exports.createSyncFileSystem = exports.createFileSystem = void 0;
const returnsTrue = () => true;
const statsNoThrowOptions = { throwIfNoEntry: false };
function createFileSystem(baseFs) {
    return {
        ...baseFs,
        ...createExtendedSyncActions(baseFs),
        promises: {
            ...baseFs.promises,
            ...createExtendedFileSystemPromiseActions(baseFs),
        },
    };
}
exports.createFileSystem = createFileSystem;
function createSyncFileSystem(baseFs) {
    return {
        ...baseFs,
        ...createExtendedSyncActions(baseFs),
    };
}
exports.createSyncFileSystem = createSyncFileSystem;
function createExtendedSyncActions(baseFs) {
    const { statSync, mkdirSync, writeFileSync, readdirSync, readFileSync, copyFileSync, dirname, join, resolve } = baseFs;
    function fileExistsSync(filePath, statFn = statSync) {
        var _a;
        try {
            return !!((_a = statFn(filePath, statsNoThrowOptions)) === null || _a === void 0 ? void 0 : _a.isFile());
        }
        catch {
            return false;
        }
    }
    function readJsonFileSync(filePath, options) {
        return JSON.parse(readFileSync(filePath, options || 'utf8'));
    }
    function directoryExistsSync(directoryPath, statFn = statSync) {
        var _a;
        try {
            return !!((_a = statFn(directoryPath, statsNoThrowOptions)) === null || _a === void 0 ? void 0 : _a.isDirectory());
        }
        catch {
            return false;
        }
    }
    function ensureDirectorySync(directoryPath) {
        try {
            mkdirSync(directoryPath);
        }
        catch (e) {
            const code = e === null || e === void 0 ? void 0 : e.code;
            if (code === 'EISDIR') {
                return;
            }
            else if (code === 'EEXIST') {
                if (directoryExistsSync(directoryPath)) {
                    return;
                }
                else {
                    throw e;
                }
            }
            else if (code === 'ENOTDIR' || !code) {
                throw e;
            }
            const parentPath = dirname(directoryPath);
            if (parentPath === directoryPath) {
                throw e;
            }
            ensureDirectorySync(parentPath);
            try {
                mkdirSync(directoryPath);
            }
            catch (e) {
                const code = e === null || e === void 0 ? void 0 : e.code;
                const isDirectoryExistsError = code === 'EISDIR' || (code === 'EEXIST' && directoryExistsSync(directoryPath));
                if (!isDirectoryExistsError) {
                    throw e;
                }
            }
        }
    }
    function populateDirectorySync(directoryPath, contents) {
        const filePaths = [];
        ensureDirectorySync(directoryPath);
        for (const [nodeName, nodeValue] of Object.entries(contents)) {
            const nodePath = join(directoryPath, nodeName);
            if (typeof nodeValue === 'string') {
                ensureDirectorySync(dirname(nodePath));
                writeFileSync(nodePath, nodeValue);
                filePaths.push(nodePath);
            }
            else {
                populateDirectorySync(nodePath, nodeValue);
            }
        }
        return filePaths;
    }
    function findFilesSync(rootDirectory, options = {}, filePaths = []) {
        const { filterFile = returnsTrue, filterDirectory = returnsTrue } = options;
        for (const item of readdirSync(rootDirectory, { withFileTypes: true })) {
            const nodePath = join(rootDirectory, item.name);
            const nodeDesc = { name: item.name, path: nodePath };
            if (item.isFile() && filterFile(nodeDesc)) {
                filePaths.push(nodePath);
            }
            else if (item.isDirectory() && filterDirectory(nodeDesc)) {
                findFilesSync(nodePath, options, filePaths);
            }
        }
        return filePaths;
    }
    function findClosestFileSync(initialDirectoryPath, fileName) {
        let currentPath = resolve(initialDirectoryPath);
        let lastPath;
        while (currentPath !== lastPath) {
            const filePath = join(currentPath, fileName);
            if (fileExistsSync(filePath)) {
                return filePath;
            }
            lastPath = currentPath;
            currentPath = dirname(currentPath);
        }
        return undefined;
    }
    function findFilesInAncestorsSync(initialDirectoryPath, fileName) {
        const filePaths = [];
        let currentPath = resolve(initialDirectoryPath);
        let lastPath;
        while (currentPath !== lastPath) {
            const filePath = join(currentPath, fileName);
            if (fileExistsSync(filePath)) {
                filePaths.push(filePath);
            }
            lastPath = currentPath;
            currentPath = dirname(currentPath);
        }
        return filePaths;
    }
    function copyDirectorySync(sourcePath, destinationPath) {
        ensureDirectorySync(destinationPath);
        for (const item of readdirSync(sourcePath, { withFileTypes: true })) {
            const sourceItemPath = join(sourcePath, item.name);
            const destinationItemPath = join(destinationPath, item.name);
            if (item.isFile()) {
                copyFileSync(sourceItemPath, destinationItemPath);
            }
            else if (item.isDirectory()) {
                copyDirectorySync(sourceItemPath, destinationItemPath);
            }
        }
    }
    return {
        fileExistsSync,
        directoryExistsSync,
        // resolve path once for recursive functions
        ensureDirectorySync: (directoryPath) => ensureDirectorySync(resolve(directoryPath)),
        populateDirectorySync: (directoryPath, contents) => populateDirectorySync(resolve(directoryPath), contents),
        findFilesSync: (rootDirectory, options) => findFilesSync(resolve(rootDirectory), options),
        copyDirectorySync: (sourcePath, destinationPath) => copyDirectorySync(resolve(sourcePath), resolve(destinationPath)),
        findClosestFileSync,
        findFilesInAncestorsSync,
        readJsonFileSync,
    };
}
exports.createExtendedSyncActions = createExtendedSyncActions;
function createAsyncFileSystem(baseFs) {
    return {
        ...baseFs,
        promises: {
            ...baseFs.promises,
            ...createExtendedFileSystemPromiseActions(baseFs),
        },
    };
}
exports.createAsyncFileSystem = createAsyncFileSystem;
function createExtendedFileSystemPromiseActions(baseFs) {
    const { dirname, resolve, join, promises: { stat, mkdir, writeFile, readdir, readFile, copyFile }, } = baseFs;
    async function fileExists(filePath, statFn = stat) {
        try {
            return (await statFn(filePath)).isFile();
        }
        catch {
            return false;
        }
    }
    async function directoryExists(directoryPath, statFn = stat) {
        try {
            return (await statFn(directoryPath)).isDirectory();
        }
        catch {
            return false;
        }
    }
    async function readJsonFile(filePath, options) {
        return JSON.parse(await readFile(filePath, options || 'utf8'));
    }
    async function ensureDirectory(directoryPath) {
        try {
            await mkdir(directoryPath);
        }
        catch (e) {
            const code = e === null || e === void 0 ? void 0 : e.code;
            if (code === 'EISDIR') {
                return;
            }
            else if (code === 'EEXIST') {
                if (await directoryExists(directoryPath)) {
                    return;
                }
                else {
                    throw e;
                }
            }
            else if (code === 'ENOTDIR' || !code) {
                throw e;
            }
            const parentPath = dirname(directoryPath);
            if (parentPath === directoryPath) {
                throw e;
            }
            await ensureDirectory(parentPath);
            try {
                await mkdir(directoryPath);
            }
            catch (e) {
                const code = e === null || e === void 0 ? void 0 : e.code;
                const isDirectoryExistsError = code === 'EISDIR' || (code === 'EEXIST' && (await directoryExists(directoryPath)));
                if (!isDirectoryExistsError) {
                    throw e;
                }
            }
        }
    }
    async function populateDirectory(directoryPath, contents) {
        const filePaths = [];
        await ensureDirectory(directoryPath);
        for (const [nodeName, nodeValue] of Object.entries(contents)) {
            const nodePath = join(directoryPath, nodeName);
            if (typeof nodeValue === 'string') {
                await ensureDirectory(dirname(nodePath));
                await writeFile(nodePath, nodeValue);
                filePaths.push(nodePath);
            }
            else {
                await populateDirectory(nodePath, nodeValue);
            }
        }
        return filePaths;
    }
    async function findFiles(rootDirectory, options = {}, filePaths = []) {
        const { filterFile = returnsTrue, filterDirectory = returnsTrue } = options;
        for (const item of await readdir(rootDirectory, { withFileTypes: true })) {
            const nodePath = join(rootDirectory, item.name);
            const nodeDesc = { name: item.name, path: nodePath };
            if (item.isFile() && filterFile(nodeDesc)) {
                filePaths.push(nodePath);
            }
            else if (item.isDirectory() && filterDirectory(nodeDesc)) {
                await findFiles(nodePath, options, filePaths);
            }
        }
        return filePaths;
    }
    async function findClosestFile(initialDirectoryPath, fileName) {
        let currentPath = resolve(initialDirectoryPath);
        let lastPath;
        while (currentPath !== lastPath) {
            const filePath = join(currentPath, fileName);
            if (await fileExists(filePath)) {
                return filePath;
            }
            lastPath = currentPath;
            currentPath = dirname(currentPath);
        }
        return undefined;
    }
    async function findFilesInAncestors(initialDirectoryPath, fileName) {
        const filePaths = [];
        let currentPath = resolve(initialDirectoryPath);
        let lastPath;
        while (currentPath !== lastPath) {
            const filePath = join(currentPath, fileName);
            if (await fileExists(filePath)) {
                filePaths.push(filePath);
            }
            lastPath = currentPath;
            currentPath = dirname(currentPath);
        }
        return filePaths;
    }
    async function copyDirectory(sourcePath, destinationPath) {
        await ensureDirectory(destinationPath);
        for (const item of await readdir(sourcePath, { withFileTypes: true })) {
            const sourceItemPath = join(sourcePath, item.name);
            const destinationItemPath = join(destinationPath, item.name);
            if (item.isFile()) {
                await copyFile(sourceItemPath, destinationItemPath);
            }
            else if (item.isDirectory()) {
                await copyDirectory(sourceItemPath, destinationItemPath);
            }
        }
    }
    return {
        fileExists,
        directoryExists,
        ensureDirectory: (directoryPath) => ensureDirectory(resolve(directoryPath)),
        populateDirectory: (directoryPath, contents) => populateDirectory(resolve(directoryPath), contents),
        findFiles: (rootDirectory, options) => findFiles(resolve(rootDirectory), options),
        copyDirectory: (sourcePath, destinationPath) => copyDirectory(resolve(sourcePath), resolve(destinationPath)),
        findClosestFile,
        findFilesInAncestors,
        readJsonFile,
    };
}
exports.createExtendedFileSystemPromiseActions = createExtendedFileSystemPromiseActions;
//# sourceMappingURL=create-extended-api.js.map