#!/usr/bin/env node
"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const yargs_1 = __importDefault(require("yargs"));
const node_1 = require("@file-services/node");
const code_formatter_1 = require("@stylable/code-formatter");
const logger_1 = require("./logger");
const fs_1 = require("fs");
const { join } = node_1.nodeFs;
const argv = yargs_1.default
    .usage('$0 [options]')
    .option('target', {
    type: 'string',
    description: 'file or directory to format',
    alias: 'T',
    default: process.cwd(),
    defaultDescription: 'current working directory',
})
    .option('endWithNewline', {
    type: 'boolean',
    description: 'End output with newline',
    alias: 'n',
    default: false,
})
    .option('indentEmptyLines', {
    type: 'boolean',
    description: 'Keep indentation on empty lines',
    alias: 'E',
    default: false,
})
    .option('indentSize', {
    type: 'number',
    description: 'Indentation size',
    alias: 's',
    default: 4,
})
    .option('indentWithTabs', {
    type: 'boolean',
    description: 'Indent with tabs, overrides -s and -c',
    alias: 't',
    default: false,
})
    .option('maxPerserveNewlines', {
    type: 'number',
    description: 'Maximum number of line-breaks to be preserved in one chunk',
    alias: 'm',
    default: 1,
})
    .option('newlineBetweenRules', {
    type: 'boolean',
    description: 'Add a newline between CSS rules',
    alias: 'N',
    default: true,
})
    .option('perserveNewlines', {
    type: 'boolean',
    description: 'Preserve existing line-breaks',
    alias: 'p',
    default: true,
})
    .option('selectorSeparatorNewline', {
    type: 'boolean',
    description: 'Add a newline between multiple selectors',
    alias: 'L',
    default: true,
})
    .option('debug', {
    type: 'boolean',
    description: 'Enable explicit debug log (overrides --silent)',
    alias: 'D',
    default: false,
})
    .option('silent', {
    type: 'boolean',
    description: 'Will not print any messages to the log',
    alias: 'S',
    default: false,
})
    .option('require', {
    type: 'array',
    description: 'require hooks',
    string: true,
    alias: 'r',
    default: [],
})
    .alias('h', 'help')
    .alias('v', 'version')
    .help()
    .strict()
    .wrap(yargs_1.default.terminalWidth())
    .parseSync();
const { debug, endWithNewline, indentEmptyLines, indentSize, indentWithTabs, maxPerserveNewlines, newlineBetweenRules, perserveNewlines, require: requires, selectorSeparatorNewline, target, silent, } = argv;
const log = (0, logger_1.createLogger)((_, ...messages) => console.log('[Stylable code formatter]', ...messages), () => console.clear());
if (debug) {
    log('[Arguments]', argv);
}
// execute all require hooks before running the CLI build
for (const request of requires) {
    require(request);
}
function readDirectoryDeep(dirPath, fileSuffixFilter = '.st.css') {
    const files = node_1.nodeFs.readdirSync(dirPath, 'utf-8');
    let res = [];
    for (const item of files) {
        const currentlFilePath = join(dirPath, item);
        const itemStat = node_1.nodeFs.statSync(join(currentlFilePath));
        if (itemStat.isFile() && item.endsWith(fileSuffixFilter)) {
            res.push(currentlFilePath);
        }
        else if (itemStat.isDirectory()) {
            res = res.concat(readDirectoryDeep(currentlFilePath));
        }
    }
    return res;
}
function formatStylesheet(filePath) {
    const fileContent = node_1.nodeFs.readFileSync(filePath, 'utf-8');
    const newText = (0, code_formatter_1.getDocumentFormatting)(fileContent, { start: 0, end: fileContent.length }, {
        end_with_newline: endWithNewline,
        indent_empty_lines: indentEmptyLines,
        indent_size: indentSize,
        indent_with_tabs: indentWithTabs,
        max_preserve_newlines: maxPerserveNewlines,
        newline_between_rules: newlineBetweenRules,
        preserve_newlines: perserveNewlines,
        selector_separator_newline: selectorSeparatorNewline,
    });
    if (newText.length) {
        (0, fs_1.writeFileSync)(filePath, newText);
        if (!silent || debug) {
            log(`Formatted: ${filePath}`);
        }
    }
    else if (debug) {
        log(`No formatting required for: ${filePath}`);
    }
}
if (debug) {
    log('Starting code formatting');
}
const formatPathStats = node_1.nodeFs.statSync(target);
if (formatPathStats.isFile()) {
    if (target.endsWith('.st.css')) {
        formatStylesheet(target);
    }
    else {
        throw new Error('cannot format file, not a Stylable stylesheet (.st.css)');
    }
}
else if (formatPathStats.isDirectory()) {
    const stylesheets = readDirectoryDeep(target);
    if (stylesheets.length) {
        for (const stylesheet of stylesheets) {
            formatStylesheet(stylesheet);
        }
    }
    else {
        throw new Error('cannot find any Stylable stylesheets (.st.css) in directory: ' + target);
    }
}
if (debug) {
    log('All code formatting complete');
}
//# sourceMappingURL=code-format.js.map