"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ComponentMetadataBuilder = void 0;
const path_1 = require("path");
const NODE_MODULES_FOLDER = `/node_modules/`;
class ComponentMetadataBuilder {
    constructor(context, name, version) {
        this.context = context;
        this.output = {
            version,
            name,
            fs: {},
            components: {},
            packages: {},
        };
    }
    hasPackages() {
        return Object.keys(this.output.packages).length;
    }
    build() {
        this.validate();
        return this.output;
    }
    addSource(resourcePath, content, metadata) {
        const local = this.localResourcePath(resourcePath);
        this.output.fs[local] = {
            metadata,
            content,
        };
        this.output.packages[this.resourcePackageName(local)] = this.packageRootPath(local);
    }
    addComponent(stylesheetPath, localComponentConfig, namespace) {
        const componentConfig = cloneObject(localComponentConfig);
        const componentDir = (0, path_1.dirname)(stylesheetPath);
        const variantsPath = componentConfig.variantsPath;
        componentConfig.namespace = namespace;
        componentConfig.stylesheetPath = this.localResourcePath(stylesheetPath);
        if (variantsPath) {
            componentConfig.variantsPath = this.localResourcePath((0, path_1.join)(componentDir, variantsPath));
            if (componentConfig.presets) {
                for (const preset of componentConfig.presets) {
                    preset.path = this.localResourcePath((0, path_1.join)(componentDir, variantsPath, preset.path));
                }
            }
        }
        if (!componentConfig.id) {
            throw new Error(`Invalid component config for resource: ${stylesheetPath}. Missing {id}`);
        }
        if (this.output.components[componentConfig.id]) {
            throw new Error(`Duplicate Component ID: ${componentConfig.id}`);
        }
        this.output.components[componentConfig.id] = componentConfig;
    }
    addComponentSnapshot(id, snapshot) {
        const componentConfig = this.output.components[id];
        componentConfig.snapshots = componentConfig.snapshots || [];
        if (Array.isArray(snapshot)) {
            componentConfig.snapshots.push(...snapshot);
        }
        else {
            componentConfig.snapshots.push(snapshot);
        }
    }
    createIndex() {
        const indexPath = (0, path_1.join)(this.context, 'index.st.css');
        const namespace = this.output.name + '-index';
        if (this.output.fs[indexPath]) {
            throw new Error('Duplicate index');
        }
        let maxDepth = 0;
        const source = Object.keys(this.output.components).reduce((source, name) => {
            const { stylesheetPath } = this.output.components[name];
            const { depth } = this.output.fs[stylesheetPath].metadata;
            const from = stylesheetPath;
            maxDepth = Math.max(maxDepth, depth);
            return source + `:import {-st-from: "${from}"; -st-default: ${name}} .root ${name}{}\n`;
        }, '');
        this.addSource(indexPath, source, { namespace, depth: maxDepth });
    }
    validate() {
        const errors = [];
        for (const component in this.output.components) {
            for (const preset of this.output.components[component].presets || []) {
                if (!this.output.fs[preset.path]) {
                    errors.push(`Missing variant for preset ${preset.path}`);
                }
            }
        }
        if (errors.length) {
            throw new Error('Invalid metadata output:\n' + errors.join('\n'));
        }
    }
    packageRootPath(localResourcePath) {
        const i = localResourcePath.lastIndexOf(NODE_MODULES_FOLDER);
        if (i !== -1) {
            return (localResourcePath.slice(0, i + NODE_MODULES_FOLDER.length) +
                this.resourcePackageName(localResourcePath));
        }
        else {
            return '/' + this.output.name;
        }
    }
    localResourcePath(resourcePath) {
        return '/' + this.output.name + normPath(resourcePath, this.context);
    }
    resourcePackageName(localResourcePath) {
        const i = localResourcePath.lastIndexOf(NODE_MODULES_FOLDER);
        if (i !== -1) {
            const [packageName, subName] = localResourcePath
                .slice(i + NODE_MODULES_FOLDER.length)
                .split('/');
            if (packageName.startsWith('@')) {
                return packageName + '/' + subName;
            }
            return packageName;
        }
        else {
            return this.output.name;
        }
    }
}
exports.ComponentMetadataBuilder = ComponentMetadataBuilder;
function normPath(resource, context = '') {
    let norm = resource.replace(context, '');
    if (norm === resource) {
        throw new Error(`cannot use resource "${resource}" outside the context "${context}"`);
    }
    norm = norm.replace(/\\/g, '/');
    return norm.startsWith('/') ? norm : `/${norm}`;
}
function cloneObject(obj) {
    const clone = (Array.isArray(obj) ? [] : {});
    for (const i in obj) {
        const v = obj[i];
        if (v && typeof v === 'object') {
            clone[i] = cloneObject(v);
        }
        else {
            clone[i] = v;
        }
    }
    return clone;
}
//# sourceMappingURL=component-metadata-builder.js.map