"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.StylableMetadataPlugin = void 0;
const node_1 = require("@stylable/node");
const path_1 = require("path");
const compile_as_entry_1 = require("./compile-as-entry");
const component_metadata_builder_1 = require("./component-metadata-builder");
const webpack_plugin_1 = require("@stylable/webpack-plugin");
const hash_content_util_1 = require("./hash-content-util");
class StylableMetadataPlugin {
    constructor(options) {
        this.options = options;
    }
    apply(compiler) {
        compiler.hooks.thisCompilation.tap('StylableMetadataPlugin', (compilation) => {
            compilation.hooks.processAssets.tapPromise('StylableMetadataPlugin', async () => {
                await this.createMetadataAssets(compilation);
            });
        });
    }
    loadComponentConfig(compilation, component) {
        return this.loadJSON(compilation.inputFileSystem, component.resource.replace(/\.[^.]+$/, this.options.configExtension || '.component.json'));
    }
    loadJSON(fs, resource) {
        try {
            return JSON.parse(fs.readFileSync(resource).toString());
        }
        catch (e) {
            if (e instanceof SyntaxError) {
                throw new SyntaxError(`${e.message} in ${resource}`);
            }
            return null;
        }
    }
    async createMetadataAssets(compilation) {
        var _a, _b;
        const stylableModules = (0, webpack_plugin_1.uniqueFilterMap)(compilation.modules, (m) => (0, webpack_plugin_1.isStylableModule)(m) ? m : null);
        const builder = new component_metadata_builder_1.ComponentMetadataBuilder(this.options.context || compilation.compiler.options.context || process.cwd(), this.options.name, this.options.version);
        const getViewModule = (0, webpack_plugin_1.getCSSViewModuleWebpack)(compilation.moduleGraph);
        const stylableModulesWithData = (0, webpack_plugin_1.getStylableModules)(compilation);
        for (const module of stylableModules) {
            const namespace = (_b = (_a = stylableModulesWithData === null || stylableModulesWithData === void 0 ? void 0 : stylableModulesWithData.get(module)) === null || _a === void 0 ? void 0 : _a.namespace) !== null && _b !== void 0 ? _b : module.buildMeta.namespace;
            const depth = module.buildMeta.stylable.depth;
            const resource = this.options.normalizeModulePath
                ? this.options.normalizeModulePath(module.resource, builder)
                : module.resource;
            builder.addSource(resource, compilation.inputFileSystem.readFileSync(resource).toString(), { namespace, depth });
            const component = getViewModule(module);
            if (!component || !component.context) {
                continue;
            }
            const componentConfig = this.loadComponentConfig(compilation, component);
            if (!componentConfig) {
                continue;
            }
            builder.addComponent(resource, componentConfig, namespace);
            this.handleVariants(componentConfig, (0, path_1.dirname)(resource), compilation, builder, namespace, depth);
            if (this.options.renderSnapshot) {
                const source = await (0, compile_as_entry_1.compileAsEntry)(compilation, component.context, component.resource, []);
                const componentModule = (0, compile_as_entry_1.exec)(source, component.resource, component.context);
                const html = this.options.renderSnapshot(componentModule, component, componentConfig);
                builder.addComponentSnapshot(componentConfig.id, html);
            }
        }
        if (builder.hasPackages()) {
            builder.createIndex();
            const jsonMode = !this.options.mode || this.options.mode === 'json';
            const jsonSource = JSON.stringify(builder.build(), null, 2);
            let fileContent = jsonSource;
            switch (this.options.mode) {
                case 'cjs':
                    fileContent = `module.exports = ${fileContent}`;
                    break;
                case 'amd:static':
                    fileContent = `define(${fileContent});`;
                    break;
                case 'amd:factory':
                    fileContent = `define(() => { return ${fileContent}; });`;
                    break;
            }
            const fileName = `${this.options.name}${this.options.useContentHashFileName
                ? `.${(0, hash_content_util_1.hashContent)(fileContent, this.options.contentHashLength)}`
                : ''}.metadata.json${!jsonMode ? '.js' : ''}`;
            compilation.emitAsset(fileName, new compilation.compiler.webpack.sources.RawSource(fileContent, false));
        }
    }
    handleVariants(componentConfig, componentDir, compilation, builder, namespace, depth) {
        if (componentConfig.variantsPath) {
            const variantsDir = (0, path_1.join)(componentDir, componentConfig.variantsPath);
            const { result: variants, errors } = (0, node_1.findFiles)(compilation.inputFileSystem, path_1.join, path_1.relative, variantsDir, '.st.css', new Set(), true);
            if (errors.length) {
                throw new Error(`Error while reading variants for: ${componentConfig.id} in ${variantsDir}\nOriginal Errors:\n${errors}`);
            }
            variants.forEach((name) => {
                if (!name.match(/\.st\.css/)) {
                    return;
                }
                const variantPath = (0, path_1.join)(variantsDir, name);
                let content;
                try {
                    content = compilation.inputFileSystem
                        .readFileSync(variantPath)
                        .toString();
                }
                catch (e) {
                    throw new Error(`Error while reading variant: ${variantPath}\nOriginal Error:\n${e}`);
                }
                if (name.includes('_')) {
                    throw new Error(`Error variant name or folder cannot contain "_" found in: ${name}`);
                }
                builder.addSource(variantPath, content, {
                    namespace: name.replace(/\\/g, '/').replace(/\//g, '_').replace('.st.css', '') +
                        '-' +
                        namespace,
                    variant: true,
                    depth,
                });
            });
        }
    }
}
exports.StylableMetadataPlugin = StylableMetadataPlugin;
//# sourceMappingURL=stylable-metadata-plugin.js.map