# Board thumbnails optimization research and desgin

## Product needs
We want the ability to render all the boards existing in a project in the same page, to serve as thumbnails.
We want the render to be fast, and the product to work fast while rendering.
We want the components to react to changes made from the product/file system but we don't have the components to be "live" (respond to click events, change states and so on). 

## Problems with the current approach
Currently, when trying to render **many**  boards simultan, we face both memory and performance issues.

### Memory
With the existing solution, each iframe is a `preview` environment, which means that each iframe stores the entire module system which is required to evaluate the transpiled user board. 
This is a massive overhead of memory when speaking about large amounts of iframes simultaniously.

### Performance
The rendering time of a component in an iframe is impacted by 3 things:
 1. Code execution - all the user's source code needs to be transpiled into commonjs modules, so that our module system will be able to evaluate them. 
     For the user's ts/js files, if they are not already commonjs files, we transform them to commonjs
     For the user's style files we also create a "commonjs" module, which upon evaluation adds a stylesheet representing the source file, compiled to a css output.
  2. Communication - Since the `preview` environment is an engine environment, it comes with some network overhead, when also loading the preview environment bundle, and evaluating it, as well as loading from the `processing` environment all the files (with contents) which are required for the module system.
  3. Code evaluation - currently, for each iframe, we need to evaluate **all** the code required for the board module to render. This means that we re-evaluate all node_module contents as well as all the user files in each iframe, which is costly and in "heavy" components might take up to 5 seconds.

## Solution
The general solution is to not have dynamic content as the content of the iframe. This means that we don't want to evaluate **any** js file when previewing a board. 
We will want to have a static html which will load fast, and represent the component, after it loaded for the first time (without taking into consideration that the component might change it's view after some time as a result of suspense, hooks, timeouts and so on).
To be able to achieve that we have 2 solutions which were reviewed.
### Cache the iframe html, and de-facto cleanup the js side effect
When we want to render a component, we will run the same flow as today, but in the end of it, will return to the `main` environment the html resulted from the rendering of the component. 
The `main` environment then will drom the `preview` environment and replace it with a different iframe, which will be the result of the previous rendering.
The `main` environment will know for each item in that cache when it needs to revalidate (resource), and while the board or any of it's components didn't change, it will display that static html, otherwise, will do the same process like it did when rendering the component for the first time.

##### Pros
- doesn't require a lot of effort
- each following render of a component will be very fast, and light in memory, which will cause smoother behavior then what we have now
##### Cons
- we still re-evaluate for each board the entire module system, so each component, when redered for the first time, or after each change, would still need to retrieve all the module system and evaaluate all the code.
- This will cause the experience of using the product when it loads to be not optimal, and depend on evaluation time.

### One preview to render them all
We will create a new environment `thumbnail` environment, which will be in-charge of code evaluation and management of the module system.
It will expose a service, when upon a demand will render the requested component in a dedicated (hidden) iframe. After it finished, it will return the static html of the component to `main`, which in it's turn will create an iframe with the html returned.
Once another request comes to that service, it will need to know how to clean all the side-effects caused by the evaluation of the code, and which files caused them, so that for next render we would be able to re-evaluate them, if needed.
like the first solution, `main` environment will decide if a specific board/component should re-render.

##### Pros
- This approach handles all pain points we have for rendering components. it will not re-evaluate code if not needed, which will cause **every** evaluation (except the first one done in the product) to be blazing fast.
- It will also have a light memory result, which will allow us to have as many iframes we want
- It will reduce the network overhead for retrieving all the compiled files required for the module system in the iframe
##### Cons
- The biggest con here is the uncertenty when it comes to clean up side-effects. we need to make sure we can intercept all of them, clean then up and attach each side-effect with the file responsible for creating it.

##### Side effects
When speaking about side-effects in the browser, it can come in various ways.

- document can be manupilated
- head can be manipulated
- global event listeners might be registered
- promises might not be resolved
and more

After a few days of POCing the 2nd approach, I am glad to say it seems that we are able to intercept all the side-effects are able to be intercepted, and also get cleaned up. This is done using [zone.js](https://www.npmjs.com/package/zone.js), so we will currently go forward with solution #2