# Theming

Currently we only have the light theme, but we might add a dark theme in the future. Themes differ only in color, other properties such as fonts and sizes should be specified in the component's base stylesheet.

Theming example:

```css
/* File: main/components/checkbox/checkbox-light-theme.st.css */

:import {
    -st-from: '../../themes/light-theme.st.css';
    -st-default: LightTheme;
}

:import {
    -st-from: '../../themes/colors.st.css';
    -st-named: white, purple;
}

:import {
    -st-from: './checkbox.st.css';
    -st-default: Checkbox;
}

@st-scope LightTheme {
    Checkbox {
        background-color: value(white);
        border-color: value(purple, 4);
    }
}
```

`LightTheme` imported from `light-theme.st.css` is simply a unique class name applied to the body element when the light theme is selected. Everything in `@st-scope LightTheme { }` becomes a descendant selector of that class name. To toggle between themes we change the body class name.

`colors.st.css` is not specific to any theme, but we might have to create another color file for the dark theme, if for example it needs a different hue of purple.

```tsx
// File: main/components/checkbox/checkbox.tsx

import { classes, style } from './checkbox.st.css';

// We need to import the theme file to include it in the compiled CSS.
// If there was a dark theme as well, we'd have a second import here.
import './checkbox-light-theme.st.css';

export const Checkbox = props => (
    <div className={style(classes.root, props.className)} />
);
```

Generally there's one theme file per component (e.g. `checkbox-light-theme.st.css`). But for small features it could be more convenient to use a single theme file for the entire feature, in which case it should be imported in `<feature-name>.main.env.ts`.

## Style Variants

Some components such as buttons have several style variants within a single theme. Regular, primary, or alert button, as well as small, medium, or large button. Instead of duplicating styles for each instance of the button we can define reusable variants:

```css
/* File: main/components/button/button-variants.st.css */

:import {
    -st-from: './button.st.css';
    -st-default: Button;
}

.primaryButton {
    -st-extends: Button;
    /* This variant should have a different background color,
       but we cannot specify it here, since the color depends
       on the theme. Instead we specify it in the theme file.
    */
}

.smallButton {
    -st-extends: Button;
    font-size: smaller;
    border-radius: 2px;
}
```

```css
/* File: main/components/button/button-light-theme.st.css */

:import {
    -st-from: '../../themes/light-theme.st.css';
    -st-default: LightTheme;
}

:import {
    -st-from: '../../themes/colors.st.css';
    -st-named: purple;
}

:import {
    -st-from: './button.st.css';
    -st-default: Button;
}

:import {
    -st-from: './button-variants.st.css';
    -st-named: primaryButton;
}

@st-scope LightTheme {
    Button { /* . . . */ }

    .primaryButton {
        background-color: value(purple, 9);
    }
}
```

There are two ways to apply a variant to an instance of the button. Using JavaScript:

```tsx
/* File: dialog.tsx */

import { classes } from './dialog.st.css';
import { classes as buttonVariants } from '@wixc3/codux-components/main/components/button/button-variants.st.css';

<Button className={style(classes.submitButton, buttonVariants.primaryButton)}>OK</Button>
```

Or using CSS:

```tsx
/* File: dialog.tsx */

import { classes } from './dialog.st.css';

<Button className={classes.submitButton}>OK</Button>
```

```css
/* File: dialog.st.css */

:import {
    -st-from: '@wixc3/codux-components/main/components/button/button-variants.st.css';
    -st-named: primaryButton;
}

.submitButton {
    -st-extends: primaryButton;
}
```

`-st-extends` here is a little tricky, it doesn't work the same as when extending the `root` of the button's base stylesheet. Extending a regular class name (such as `primaryButton` here) alters the value of `classes.submitButton` in JavaScript. Without `-st-extends` it would be `dialog1234__submitButton`, but with `-stextends` it's `dialog1234__submitButton buttonvariants1234__primaryButton`. The result is the same as if we added the class in JavaScript.

Currently Stylable doesn't allow to provide mutliple classes: `-st-extends: primaryButton, smallButton`, as a workaround we can use the JavaScript approach.

## Appendix

### What about semantic variables?

Wouldn't it be so much better if instead of defining a mapping between each element and its color: `Checkbox:checked:hover { color: value(pink, 7); }`, we used semantic variables like `--surface-primary-hover-color`.

A theme would be a single file with a list of variable assignments:

```css
.root {
    --surface-primary-color: value(pink, 9);
    --surface-primary-hover-color: value(pink, 11);
    /* ... more ... */
}
```

And a single `checkbox.st.css` would support any number of themes:

```css
:import {
    -st-from: '@wixc3/codux-components/theme-vars.st.css';
    -st-named: --surface-primary-hover-color;
}

.root:checked:hover {
    background-color: var(--surface-primary-hover-color);
}
```

This would be great indeed. But creating a semantic design system requires *a lot* of work from the designer. Ideally two themes should be created in parallel to ensure that the system actually works.

Trying to cobble something together without designer's involvement would most likely result in a mess of variables without any system. Developers would constantly have to ask themselves questions like "what variable should I use for the checked state of the checkbox, is it `--primary-color`, or `--surface-background-inverted`, or `--input-accent-color` or one of these other five variables that seem relevant." This would get even more messy if there was an alternative checkbox style requiring different colors.

### Using mixins for theming

This is an extension of the technique described in the main section. For a full example see [#2309](https://github.com/wixplosives/codux/pull/2309).

```css
/* File: checkbox-color-mixin.st.css */

:import {
    -st-from: './checkbox.st.css';
    -st-default: Checkbox;
}

:vars {
    background: none;
    border: none;
    checked-background: none;
    invalid-checked-border: none;
    /* ... more ... */
}

.checkboxColors {
    -st-extends: Checkbox;
    background-color: value(background);
    border-color: value(border);
}

.checkboxColors:checked {
    background-color: value(checked-background);
}

.checkboxColors:invalid:checked {
    border-color: value(invalid-checked-border);
}
```

```css
/* File: checkbox-light-theme.st.css */

:import {
    -st-from: '../../themes/light-theme.st.css';
    -st-default: LightTheme;
}

:import {
    -st-from: './checkbox.st.css';
    -st-default: Checkbox;
}

:import {
    -st-from: './checkbox-variants.st.css';
    -st-named: secondaryCheckbox;
}

:import {
    -st-from: './checkbox-color-mixin.st.css';
    -st-named: checkboxColors;
}

@st-scope LightTheme {
    Checkbox {
        -st-mixin: checkboxColors(
            background              value(white),
            border                  value(purple, 4),
            checked-background      value(purple, 9),
            invalid-checked-border  value(red, 3),
            /* ... more ... */
        );
    }

    .secondaryCheckbox {
        -st-mixin: checkboxColors(
            background              value(white),
            border                  value(purple, 4),
            checked-background      value(purple, 16),
            invalid-checked-border  value(red, 3),
            /* ... more ... */
        );
    }
}
```

There are a few problems with this approach:

1. For every variant you have to provide the full list of variables, even if most of them could be inherited from the default variant.
2. After adding or removing a variable you have to update every place where the mixin is used.
3. Forgetting to provide one of the mixin's variables or providing an unsupported variable doesn't trigger an error.
3. `.secondaryCheckbox { invalid-checked-border value(red, 3) }` although somewhat shorter, is semantically equivalent to `.secondaryCheckbox:invalid:checked { border-color: value(red, 3); }`. But the latter doesn't introduce an additional layer of indirection.
