"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.declareComEmitter = exports.Communication = void 0;
const errors_1 = require("./errors");
const helpers_1 = require("./helpers");
const types_1 = require("./types");
const symbols_1 = require("../symbols");
const helpers_2 = require("../helpers");
const patterns_1 = require("@wixc3/patterns");
const base_host_1 = require("./hosts/base-host");
const ws_client_host_1 = require("./hosts/ws-client-host");
const message_types_1 = require("./message-types");
const promise_assist_1 = require("promise-assist");
/**
 * Manages all API registrations and message forwarding
 * in each execution context.
 */
class Communication {
    constructor(host, id, topology = {}, resolvedContexts = {}, isServer = false, options) {
        this.topology = topology;
        this.resolvedContexts = resolvedContexts;
        this.isServer = isServer;
        this.idsCounter = new helpers_1.MultiCounter();
        this.disposing = false;
        this.callbackTimeout = 60000 * 5; // 5 minutes
        this.slowThreshold = 5000; // 5 seconds
        this.callbacks = {};
        this.pendingEnvs = new patterns_1.SetMultiMap();
        this.pendingMessages = new patterns_1.SetMultiMap();
        this.handlers = new Map();
        this.eventDispatchers = {};
        this.apis = {};
        this.apisOverrides = {};
        this.readyEnvs = new Set();
        this.environments = {};
        this.messageHandlers = new WeakMap();
        this.disposeListeners = new Set();
        this.callbackToEnvMapping = new Map();
        // the any is here because issue with typescript and union types Worker | Window have different message handlers
        this.handleEvent = ({ data }) => {
            this.handleMessage(data).catch(errors_1.reportError);
        };
        this.options = { warnOnSlow: false, publicPath: '', connectedEnvironments: {}, ...options };
        this.rootEnvId = id;
        this.rootEnvName = id.split('/')[0];
        this.registerMessageHandler(host);
        this.registerEnv(id, host);
        this.environments['*'] = { id, host };
        this.post(this.getPostEndpoint(host), {
            type: 'ready',
            from: id,
            to: '*',
            origin: id,
            forwardingChain: [],
        });
        for (const [id, envEntry] of Object.entries(this.options.connectedEnvironments)) {
            if (envEntry.registerMessageHandler) {
                this.registerMessageHandler(envEntry.host);
            }
            this.environments[id] = envEntry;
        }
    }
    /**
     * Registers environments that spawned in the same execution context as the root environment.
     */
    registerEnv(id, host) {
        if (!this.environments[id]) {
            this.environments[id] = { id, host };
        }
        else {
            throw new Error((0, errors_1.DUPLICATE_REGISTER)(id, 'Environment'));
        }
    }
    /**
     * Registers local api implementation of the remote service.
     */
    registerAPI({ id }, api) {
        if (!this.apis[id]) {
            this.apis[id] = api;
            this.applyApiDirectives(id, api);
            return api;
        }
        else {
            throw new Error((0, errors_1.DUPLICATE_REGISTER)(id, 'RemoteService'));
        }
    }
    getEnvironmentContext(endPoint) {
        return this.resolvedContexts[endPoint.env];
    }
    getEnvironmentInstanceId(envName, endpointType) {
        return endpointType === 'single' ? envName : this.generateEnvInstanceID(envName);
    }
    getPublicPath() {
        return this.options.publicPath;
    }
    setTopology(envName, envUrl) {
        this.topology[envName] = envUrl;
    }
    subscribeToEnvironmentDispose(handler) {
        this.disposeListeners.add(handler);
    }
    unsubscribeToEnvironmentDispose(handler) {
        this.disposeListeners.delete(handler);
    }
    /**
     * Creates a Proxy for a remote service api.
     */
    apiProxy(instanceToken, { id: api }, serviceComConfig = {}) {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-return
        return new Proxy(Object.create(null), {
            get: (obj, method) => {
                if (typeof method === 'string') {
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-member-access
                    let runtimeMethod = obj[method];
                    if (!runtimeMethod) {
                        runtimeMethod = async (...args) => this.callMethod((await instanceToken).id, api, method, args, this.rootEnvId, serviceComConfig, []);
                        // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-member-access
                        obj[method] = runtimeMethod;
                    }
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                    return runtimeMethod;
                }
            },
        });
    }
    /**
     * Add local handle event listener to Target.
     */
    registerMessageHandler(target) {
        var _a;
        const envTarget = (_a = target.parent) !== null && _a !== void 0 ? _a : target;
        const onTargetMessage = ({ data }) => {
            if ((0, message_types_1.isMessage)(data)) {
                if (!this.environments[data.from]) {
                    this.registerEnv(data.from, envTarget);
                }
                if (!this.environments[data.origin]) {
                    this.registerEnv(data.origin, envTarget);
                }
                this.handleEvent({ data });
            }
        };
        target.addEventListener('message', onTargetMessage, true);
        this.messageHandlers.set(target, onTargetMessage);
    }
    /**
     * Remove local handle event listener to Target.
     */
    removeMessageHandler(target) {
        const messageHandler = this.messageHandlers.get(target);
        if (messageHandler) {
            target.removeEventListener('message', messageHandler, true);
        }
    }
    /**
     * Generate client id for newly spawned environment.
     */
    generateEnvInstanceID(name) {
        return this.idsCounter.next(`${name}/`);
    }
    /**
     * Calls a remote method in any opened environment.
     */
    callMethod(envId, api, method, args, origin, serviceComConfig, forwardingChain) {
        return new Promise((res, rej) => {
            var _a, _b;
            const callbackId = !((_a = serviceComConfig[method]) === null || _a === void 0 ? void 0 : _a.emitOnly) ? this.idsCounter.next('c') : undefined;
            if (this.isListenCall(args) || ((_b = serviceComConfig[method]) === null || _b === void 0 ? void 0 : _b.removeAllListeners)) {
                this.addOrRemoveListener(envId, api, method, callbackId, origin, serviceComConfig, args[0], forwardingChain, res, rej);
            }
            else {
                const message = {
                    to: envId,
                    from: this.rootEnvId,
                    type: 'call',
                    data: { api, method, args },
                    callbackId,
                    origin,
                    forwardingChain,
                };
                this.callWithCallback(envId, message, callbackId, res, rej);
            }
        });
    }
    /**
     * handles Communication incoming message.
     */
    async handleMessage(message) {
        if (message.type === 'dispose' && message.to === '*') {
            this.clearEnvironment(message.origin, this.rootEnvId);
            return;
        }
        const env = this.environments[message.to];
        if (!env) {
            this.unhandledMessage(message);
            return;
        }
        if (env.id !== this.rootEnvId) {
            await this.forwardMessage(message, env);
            return;
        }
        switch (message.type) {
            case 'call':
                await this.handleCall(message);
                break;
            case 'callback':
                this.handleCallback(message);
                break;
            case 'event':
                this.handleEventMessage(message);
                break;
            case 'listen':
                await this.handleListen(message);
                break;
            case 'unlisten':
                await this.handleUnListen(message);
                break;
            case 'ready':
                this.handleReady(message);
                break;
            case 'dispose':
                if (message.from !== this.rootEnvId) {
                    this.clearEnvironment(message.origin, message.from);
                }
                break;
            default:
                break;
        }
    }
    /**
     * Dispose the Communication and stop listening to messages.
     */
    dispose() {
        this.disposing = true;
        for (const { host, id } of Object.values(this.environments)) {
            if (host instanceof ws_client_host_1.WsClientHost) {
                host.subscribers.clear();
                host.dispose();
            }
            this.removeMessageHandler(host);
            this.locallyClear(id);
        }
        this.locallyClear(this.rootEnvId);
        this.disposeListeners.clear();
        for (const [id, { timerId }] of Object.entries(this.callbacks)) {
            clearTimeout(timerId);
            delete this.callbacks[id];
        }
    }
    getEnvironmentId() {
        return this.rootEnvId;
    }
    getEnvironmentName() {
        return this.rootEnvName;
    }
    getEnvironmentHost(envName) {
        var _a;
        return (_a = this.environments[envName]) === null || _a === void 0 ? void 0 : _a.host;
    }
    /**
     * Gets a list of all registered environment instance ids.
     * @returns the list of instance ids.
     */
    getRegisteredEnvironmentInstances() {
        return Object.keys(this.environments).filter((id) => id !== '*');
    }
    parseHandlerId(handlerId, prelude) {
        const [api, method] = handlerId.slice(prelude.length).split('@');
        return {
            api,
            method,
        };
    }
    reconnectHandler(instanceId, handlerId) {
        return new Promise((res, rej) => {
            const message = {
                to: instanceId,
                from: this.rootEnvId,
                type: 'listen',
                data: this.parseHandlerId(handlerId, this.createHandlerIdPrefix({ from: this.rootEnvId, to: instanceId })),
                callbackId: this.idsCounter.next('c'),
                origin: this.rootEnvId,
                handlerId,
                forwardingChain: [],
            };
            this.createCallbackRecord(message, message.callbackId, res, rej);
            this.sendTo(instanceId, message);
        });
    }
    async reconnectHandlers(instanceId) {
        const handlerPrefix = this.createHandlerIdPrefix({ from: this.rootEnvId, to: instanceId });
        for (const handlerId of this.handlers.keys()) {
            if (handlerId.startsWith(handlerPrefix)) {
                await this.reconnectHandler(instanceId, handlerId);
            }
        }
    }
    createHandlerIdPrefix({ from, to }) {
        return `${from}__${to}_`;
    }
    applyApiDirectives(id, api) {
        const serviceConfig = api[symbols_1.SERVICE_CONFIG];
        if (serviceConfig) {
            this.apisOverrides[id] = {};
            for (const methodName of Object.keys(serviceConfig)) {
                const config = serviceConfig[methodName](api);
                if (config.proxyFunction) {
                    this.apisOverrides[id][methodName] = config.proxyFunction;
                }
            }
        }
    }
    envReady(instanceId) {
        const { promise, resolve } = (0, promise_assist_1.deferred)();
        if (this.readyEnvs.has(instanceId)) {
            this.handleReady({ from: instanceId });
            resolve();
        }
        else {
            this.pendingEnvs.add(instanceId, () => resolve());
        }
        return promise;
    }
    clearEnvironment(instanceId, from = instanceId, emitRemote = true) {
        const connectedEnvs = Object.keys(this.options.connectedEnvironments);
        if (emitRemote && (this.readyEnvs.has(instanceId) || connectedEnvs.includes(instanceId))) {
            for (const env of [...this.readyEnvs, ...connectedEnvs]) {
                if (![instanceId, from, this.rootEnvId].includes(env)) {
                    this.sendTo(env, {
                        type: 'dispose',
                        from: this.rootEnvId,
                        to: env,
                        origin: instanceId,
                        forwardingChain: [],
                    });
                }
            }
        }
        this.locallyClear(instanceId);
    }
    locallyClear(instanceId) {
        var _a;
        this.readyEnvs.delete(instanceId);
        this.pendingMessages.deleteKey(instanceId);
        this.pendingEnvs.deleteKey(instanceId);
        delete this.environments[instanceId];
        for (const [callbackId, env] of (_a = this.callbackToEnvMapping.entries()) !== null && _a !== void 0 ? _a : []) {
            if (env === instanceId && this.callbacks[callbackId]) {
                const { timerId, reject } = this.callbacks[callbackId];
                clearTimeout(timerId);
                reject(new Error((0, errors_1.ENV_DISCONNECTED)(instanceId)));
                delete this.callbacks[callbackId];
                this.callbackToEnvMapping.delete(callbackId);
            }
        }
        for (const dispose of this.disposeListeners) {
            dispose(instanceId);
        }
    }
    async forwardMessage(message, env) {
        const responseMessage = {
            from: message.to,
            to: message.from,
            callbackId: message.callbackId,
            origin: message.to,
            type: 'callback',
            forwardingChain: [this.rootEnvId],
        };
        if (message.forwardingChain.indexOf(this.rootEnvId) > -1) {
            this.sendTo(message.from, {
                ...responseMessage,
                error: new Error(`cannot reach environment '${message.to}' from '${message.from}' since it's stuck in circular messaging loop`),
            });
            return;
        }
        message.forwardingChain = [...message.forwardingChain, this.rootEnvId];
        if (message.type === 'call') {
            try {
                const data = await this.callMethod(env.id, message.data.api, message.data.method, message.data.args, message.origin, {}, message.forwardingChain);
                if (message.callbackId) {
                    this.sendTo(message.from, { ...responseMessage, data });
                }
            }
            catch (error) {
                if (message.callbackId) {
                    this.sendTo(message.from, { ...responseMessage, error: (0, helpers_2.serializeError)(error) });
                }
            }
        }
        else if (message.type === 'callback' || message.type === 'dispose') {
            if (message.type === 'callback' && message.callbackId) {
                if (this.callbacks[message.callbackId]) {
                    const { resolve, timerId } = this.callbacks[message.callbackId];
                    resolve(message.data);
                    clearTimeout(timerId);
                }
                else {
                    this.sendTo(message.to, message);
                }
            }
            else {
                this.sendTo(message.to, message);
            }
        }
        else if (message.type === 'unlisten') {
            await this.forwardUnlisten(message);
        }
        else if (message.type === 'listen') {
            await this.forwardListenMessage(message);
        }
    }
    async forwardListenMessage(message) {
        const callbackId = this.idsCounter.next('c');
        const data = await new Promise((res, rej) => {
            const handlerId = message.handlerId;
            const handler = (...args) => {
                this.sendTo(message.from, {
                    to: message.from,
                    from: message.to,
                    type: 'event',
                    data: args,
                    handlerId,
                    origin: message.from,
                    forwardingChain: [],
                });
            };
            this.eventDispatchers[handlerId] = handler;
            this.addOrRemoveListener(message.to, message.data.api, message.data.method, callbackId, message.origin, { [message.data.method]: { listener: true } }, handler, message.forwardingChain, res, rej);
        });
        const replyCallback = {
            to: message.from,
            type: 'callback',
            callbackId: message.callbackId,
            from: message.to,
            origin: message.origin,
            data,
            forwardingChain: [],
        };
        this.sendTo(message.from, replyCallback);
    }
    apiCall(origin, api, method, args) {
        var _a;
        if ((_a = this.apisOverrides[api]) === null || _a === void 0 ? void 0 : _a[method]) {
            return this.apisOverrides[api][method](...[origin, ...args]);
        }
        return this.apis[api][method](...args);
    }
    unhandledMessage(_message) {
        // console.warn(
        //   `unhandledMessage at ${this.rootEnv} message:\n${JSON.stringify(message, null, 2)}`
        // )
    }
    addOrRemoveListener(envId, api, method, callbackId, origin, serviceComConfig, fn, forwardingChain, res, rej) {
        var _a, _b, _c, _d;
        const removeListenerRef = ((_a = serviceComConfig[method]) === null || _a === void 0 ? void 0 : _a.removeAllListeners) || ((_b = serviceComConfig[method]) === null || _b === void 0 ? void 0 : _b.removeListener);
        if (removeListenerRef) {
            const listenerHandlerId = this.getHandlerId(envId, api, removeListenerRef);
            const listenerHandlersBucket = this.handlers.get(listenerHandlerId);
            if (!listenerHandlersBucket) {
                res();
                return;
            }
            if ((_c = serviceComConfig[method]) === null || _c === void 0 ? void 0 : _c.removeListener) {
                listenerHandlersBucket.delete(fn);
            }
            else {
                listenerHandlersBucket.clear();
            }
            if (listenerHandlersBucket.size === 0) {
                // send remove handler call
                const message = {
                    to: envId,
                    from: this.rootEnvId,
                    type: 'unlisten',
                    data: {
                        api,
                        method,
                    },
                    handlerId: listenerHandlerId,
                    origin,
                    forwardingChain,
                };
                // sometimes the callback will never happen since target environment is already dead
                this.sendTo(envId, message);
                res();
            }
            else {
                res();
            }
        }
        else {
            if ((_d = serviceComConfig[method]) === null || _d === void 0 ? void 0 : _d.listener) {
                const handlersBucket = this.handlers.get(this.getHandlerId(envId, api, method));
                if (handlersBucket && handlersBucket.size !== 0) {
                    if (handlersBucket.has(fn)) {
                        throw new Error('Cannot add same listener instance twice ' + this.getHandlerId(envId, api, method));
                    }
                    handlersBucket.add(fn);
                    res();
                }
                else {
                    const message = {
                        to: envId,
                        from: this.rootEnvId,
                        type: 'listen',
                        data: {
                            api,
                            method,
                        },
                        handlerId: this.createHandlerRecord(envId, api, method, fn),
                        callbackId,
                        origin,
                        forwardingChain,
                    };
                    this.callWithCallback(envId, message, callbackId, res, rej);
                }
            }
            else {
                throw new Error(`cannot add listener to unconfigured method ${api} ${method}`);
            }
        }
    }
    callWithCallback(envId, message, callbackId, res, rej) {
        if (callbackId) {
            this.createCallbackRecord(message, callbackId, res, rej);
        }
        this.sendTo(envId, message);
        if (!callbackId) {
            res();
        }
    }
    sendTo(envId, message) {
        if (this.pendingEnvs.get(envId)) {
            this.pendingMessages.add(envId, () => this.post(this.resolveMessageTarget(envId), message));
        }
        else {
            this.post(this.resolveMessageTarget(envId), message);
        }
    }
    post(target, message) {
        if ((0, helpers_1.isWorkerContext)(target)) {
            target.postMessage(message);
        }
        else {
            target.postMessage(message, '*');
        }
    }
    resolveMessageTarget(envId) {
        const env = this.environments[envId] || this.environments[this.rootEnvId];
        if (!env) {
            return types_1.HOST_REMOVED;
        }
        const { host } = env;
        if (env.id !== this.rootEnvId) {
            return host;
        }
        else {
            if (host instanceof base_host_1.BaseHost) {
                return host.parent || host;
            }
            return this.getPostEndpoint(host);
        }
    }
    getPostEndpoint(target) {
        return (0, helpers_1.isWindow)(target) ? target.opener || target.parent : target;
    }
    handleEventMessage(message) {
        const handlers = this.handlers.get(message.handlerId);
        if (!handlers) {
            return;
        }
        for (const handler of handlers) {
            handler(...message.data);
        }
    }
    handleReady({ from }) {
        this.readyEnvs.add(from);
        const pendingEnvCb = this.pendingEnvs.get(from);
        if (pendingEnvCb) {
            this.pendingEnvs.deleteKey(from);
            const pendingMessages = this.pendingMessages.get(from);
            if (pendingMessages) {
                for (const postMessage of pendingMessages) {
                    postMessage();
                }
                this.pendingMessages.deleteKey(from);
            }
            for (const cb of pendingEnvCb) {
                cb();
            }
        }
    }
    async handleUnListen(message) {
        const namespacedHandlerId = message.handlerId + message.origin;
        const dispatcher = this.eventDispatchers[namespacedHandlerId];
        if (dispatcher) {
            delete this.eventDispatchers[namespacedHandlerId];
            const data = await this.apiCall(message.origin, message.data.api, message.data.method, [dispatcher]);
            if (message.callbackId) {
                this.sendTo(message.from, {
                    to: message.from,
                    from: this.rootEnvId,
                    type: 'callback',
                    data,
                    callbackId: message.callbackId,
                    origin: this.rootEnvId,
                    forwardingChain: [],
                });
            }
        }
    }
    async forwardUnlisten(message) {
        const callbackId = this.idsCounter.next('c');
        const { method, api } = this.parseHandlerId(message.handlerId, this.createHandlerIdPrefix(message));
        const data = await new Promise((res, rej) => this.addOrRemoveListener(message.to, api, message.data.method, callbackId, message.origin, {
            [message.data.method]: {
                removeListener: method,
            },
        }, this.eventDispatchers[message.handlerId], message.forwardingChain, res, rej));
        delete this.eventDispatchers[message.handlerId];
        if (message.callbackId) {
            this.sendTo(message.from, {
                to: message.from,
                from: message.to,
                type: 'callback',
                data,
                callbackId: message.callbackId,
                origin: message.to,
                forwardingChain: [],
            });
        }
    }
    async handleListen(message) {
        try {
            const namespacedHandlerId = message.handlerId + message.origin;
            const dispatcher = this.eventDispatchers[namespacedHandlerId] || this.createDispatcher(message.from, message);
            const data = await this.apiCall(message.origin, message.data.api, message.data.method, [dispatcher]);
            if (message.callbackId) {
                this.sendTo(message.from, {
                    to: message.from,
                    from: this.rootEnvId,
                    type: 'callback',
                    data,
                    callbackId: message.callbackId,
                    origin: this.rootEnvId,
                    forwardingChain: [],
                });
            }
        }
        catch (error) {
            this.sendTo(message.from, {
                to: message.from,
                from: this.rootEnvId,
                type: 'callback',
                error: (0, helpers_2.serializeError)(error),
                callbackId: message.callbackId,
                origin: this.rootEnvId,
                forwardingChain: [],
            });
        }
    }
    async handleCall(message) {
        try {
            const data = await this.apiCall(message.origin, message.data.api, message.data.method, message.data.args);
            if (message.callbackId) {
                this.sendTo(message.from, {
                    to: message.origin,
                    from: this.rootEnvId,
                    type: 'callback',
                    data,
                    callbackId: message.callbackId,
                    origin: this.rootEnvId,
                    forwardingChain: [],
                });
            }
        }
        catch (error) {
            this.sendTo(message.from, {
                to: message.from,
                from: this.rootEnvId,
                type: 'callback',
                error: (0, helpers_2.serializeError)(error),
                callbackId: message.callbackId,
                origin: this.rootEnvId,
                forwardingChain: [],
            });
        }
    }
    handleCallback(message) {
        const rec = message.callbackId ? this.callbacks[message.callbackId] : null;
        if (rec && message.error) {
            // If the error is not caught later, and logged to the console,
            // it would be nice to indicate which environment the error came from.
            // We shouldn't change the error message or the error name because
            // those could be used by error-handling code. Fortunately, we can
            // add metadata to the stack trace, and Chrome logs the entire stack
            // property to the console. Unfortunately, Firefox doesn't.
            const error = Object.assign(new Error(), message.error);
            error.stack = (0, errors_1.REMOTE_CALL_FAILED)(message.from, error.stack);
            rec.reject(error);
        }
        else if (rec) {
            rec.resolve(message.data);
        }
        else {
            // TODO: only in dev mode
            if (message.callbackId) {
                throw new Error((0, errors_1.UNKNOWN_CALLBACK_ID)(removeMessageArgs(message)));
            }
        }
    }
    createDispatcher(envId, message) {
        const namespacedHandlerId = message.handlerId + message.origin;
        return (this.eventDispatchers[namespacedHandlerId] = (...args) => {
            this.sendTo(envId, {
                to: envId,
                from: this.rootEnvId,
                type: 'event',
                data: args,
                handlerId: message.handlerId,
                origin: this.rootEnvId,
                forwardingChain: [],
            });
        });
    }
    isListenCall(args) {
        return typeof args[0] === 'function' && args.length === 1;
    }
    getHandlerId(envId, api, method) {
        return `${this.createHandlerIdPrefix({ from: this.rootEnvId, to: envId })}${api}@${method}`;
    }
    createHandlerRecord(envId, api, method, fn) {
        const handlerId = this.getHandlerId(envId, api, method);
        const handlersBucket = this.handlers.get(handlerId);
        if (!handlersBucket) {
            this.subscribeToEnvironmentDispose((disposedEnvId) => {
                if (envId === disposedEnvId) {
                    this.handlers.delete(handlerId);
                }
            });
        }
        handlersBucket ? handlersBucket.add(fn) : this.handlers.set(handlerId, new Set([fn]));
        return handlerId;
    }
    createCallbackRecord(message, callbackId, res, rej) {
        if (!this.disposing) {
            this.callbackToEnvMapping.set(callbackId, message.to);
            const resolve = (value) => {
                this.callbackToEnvMapping.delete(callbackId);
                delete this.callbacks[callbackId];
                clearTimeout(timerId);
                res(value);
            };
            const reject = (error) => {
                this.callbackToEnvMapping.delete(callbackId);
                delete this.callbacks[callbackId];
                clearTimeout(timerId);
                rej(error);
            };
            if (this.options.warnOnSlow) {
                setTimeout(() => {
                    if (this.callbacks[callbackId]) {
                        console.error((0, errors_1.CALLBACK_TIMEOUT)(callbackId, this.rootEnvId, removeMessageArgs(message)));
                    }
                }, this.slowThreshold);
            }
            const timerId = setTimeout(() => {
                reject(new Error((0, errors_1.CALLBACK_TIMEOUT)(callbackId, this.rootEnvId, removeMessageArgs(message))));
            }, this.callbackTimeout);
            this.callbacks[callbackId] = {
                timerId,
                resolve,
                reject,
            };
        }
    }
}
exports.Communication = Communication;
const removeMessageArgs = (message) => {
    message = { ...message };
    if (message.type === 'call' || message.type === 'callback') {
        const { data } = message;
        if (data && data.args) {
            data.args = undefined;
        }
    }
    return message;
};
function declareComEmitter(onMethod, offMethod, removeAll) {
    if (typeof onMethod !== 'string') {
        throw 'onMethod ref must be a string';
    }
    return {
        [onMethod]: { listener: true },
        [offMethod]: { removeListener: onMethod },
        ...(removeAll ? { [removeAll]: { removeAllListeners: onMethod } } : undefined),
    };
}
exports.declareComEmitter = declareComEmitter;
//# sourceMappingURL=communication.js.map