"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.deferredIframeInitializer = exports.iframeInitializer = exports.INSTANCE_ID_PARAM_NAME = void 0;
const window_initializer_service_1 = require("../window-initializer-service");
exports.INSTANCE_ID_PARAM_NAME = 'iframe-instance-id';
async function iframeInitializer({ communication, env, ...initializerOptions }) {
    const { initialize } = deferredIframeInitializer({ communication, env });
    const id = await initialize(initializerOptions);
    return {
        id
    };
}
exports.iframeInitializer = iframeInitializer;
function deferredIframeInitializer({ communication: com, env: { env, endpointType } }) {
    const instanceId = com.getEnvironmentInstanceId(env, endpointType);
    const envReadyPromise = com.envReady(instanceId);
    return {
        id: instanceId,
        initialize: ({ iframeElement, hashParams, src, origin }) => {
            const publicPath = com.getPublicPath();
            const startIframeParams = {
                com,
                envReadyPromise,
                instanceId,
                iframe: iframeElement,
                src: src !== null && src !== void 0 ? src : defaultHtmlSourceFactory(env, publicPath, hashParams, origin)
            };
            return startIframe(startIframeParams);
        }
    };
}
exports.deferredIframeInitializer = deferredIframeInitializer;
const cancellationTriggers = new WeakMap();
async function startIframe({ com, iframe, instanceId, src, envReadyPromise }) {
    var _a;
    if (!iframe.contentWindow) {
        throw new Error('Cannot initialize environment in a detached iframe');
    }
    (_a = cancellationTriggers.get(iframe)) === null || _a === void 0 ? void 0 : _a();
    const waitForCancel = new Promise((_, reject) => {
        cancellationTriggers.set(iframe, () => {
            reject('Cancelled environment initialization in an iframe');
        });
    });
    const cleanup = () => {
        com.clearEnvironment(instanceId);
    };
    try {
        const contentWindow = iframe.contentWindow;
        const url = new URL(src, window.location.href);
        com.registerEnv(instanceId, contentWindow);
        // pass instance id in query param for engine initialization
        // script that will run in the iframe can pick it up
        url.searchParams.set(exports.INSTANCE_ID_PARAM_NAME, instanceId);
        iframe.src = url.href;
        await Promise.race([waitForCancel, waitForLoad(iframe), envReadyPromise]);
        const api = com.apiProxy({ id: instanceId }, { id: window_initializer_service_1.WindowInitializerService.apiId }, {
            oncePageHide: {
                listener: true
            }
        });
        const postInitHref = await api.getHref();
        if (iframe.contentWindow !== contentWindow || postInitHref !== url.href) {
            throw new Error('Iframe location has changed during environment initialization');
        }
        // pagehide is used on purpose since it is more reliable than onUnload
        // onUnload does not send postMessages on cross origin
        // https://bugs.chromium.org/p/chromium/issues/detail?id=964950
        await api.oncePageHide(() => {
            cleanup();
        });
        cancellationTriggers.delete(iframe);
        return instanceId;
    }
    catch (e) {
        cleanup();
        throw e;
    }
}
const waitForLoad = (elem) => new Promise((resolve) => elem.addEventListener('load', resolve, { once: true }));
const defaultHtmlSourceFactory = (envName, publicPath = '', hashParams, origin = '') => {
    return `${origin}${publicPath}${envName}.html${location.search}${hashParams !== null && hashParams !== void 0 ? hashParams : ''}`;
};
//# sourceMappingURL=iframe.js.map