"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Feature = exports.RuntimeFeature = void 0;
const symbols_1 = require("../symbols");
const env_1 = require("./env");
const patterns_1 = require("@wixc3/patterns");
const promise_assist_1 = require("promise-assist");
const emptyDispose = { dispose: () => undefined };
/**
 * Represents a currently running feature instance.
 **/
class RuntimeFeature {
    constructor(feature, api, dependencies) {
        this.feature = feature;
        this.api = api;
        this.dependencies = dependencies;
        this.running = false;
        this.runHandlers = new patterns_1.SetMultiMap();
        this.disposeHandlers = new patterns_1.SetMultiMap();
    }
    addRunHandler(fn, envName) {
        this.runHandlers.add(envName, fn);
    }
    addOnDisposeHandler(fn, envName) {
        this.disposeHandlers.add(envName, fn);
    }
    async [symbols_1.RUN](context) {
        if (this.running) {
            return;
        }
        this.running = true;
        const runPromises = [];
        for (const envName of context.referencedEnvs) {
            for (const dep of this.feature.dependencies) {
                runPromises.push(context.runFeature(dep));
            }
            const envRunHandlers = this.runHandlers.get(envName) || [];
            for (const handler of envRunHandlers) {
                runPromises.push(handler());
            }
        }
        await Promise.all(runPromises);
    }
    async [symbols_1.DISPOSE](context) {
        const { entryEnvironment: { env: envName } } = context;
        if (this.disposing) {
            return this.disposing.promise;
        }
        this.disposing = (0, promise_assist_1.deferred)();
        for (const dep of this.feature.dependencies) {
            await context.dispose(dep);
        }
        const featureDisposeHandlers = this.disposeHandlers.get(envName) || new Set();
        for (const handler of featureDisposeHandlers) {
            await handler();
        }
        return this.disposing.resolve();
    }
}
exports.RuntimeFeature = RuntimeFeature;
/**
 * Feature instances define an area of functionality of an app.
 * Each has a unique string ID, and may expose a public API containing services and slots.
 * It also contains a list of all dependencies, meaning other Feature instances, to use their APIs.
 *
 * When calling `runEngineApp` and providing it with a Feature[], it uses the definitions
 * to initialize a `RuntimeFeature` for each.
 *
 * @typeParam ID - Unique string identity
 * @typeParam Deps - Dependencies needed by this Feature (to consume their API)
 * @typeParam API - Entities (Slots, Services, etc.) this Feature exposes for other Features to consume.
 * @typeParam EnvironmentContext - Environment-specific APIs consumed by this feature
 */
class Feature {
    /**
     * Define a new feature by providing:
     * - Unique string identifier
     * - API
     * - Dependencies (optional)
     * - Context (optional)
     *
     * This instance will be detected by engine when exported as default, as well as being used
     * by `<feature-name>.<environment-name>.env.ts` files to define the environment specific implementations.
     *
     * @example
     * ```
     * export default new Feature({
     *   id: 'my-feature',
     *   api: { ... },
     * })
     * ```
     */
    constructor(def) {
        /**
         * References `this` without the exact types of `Deps` (making them a generic `Feature[]`).
         * We use `someFeature.asEntity` instead of `someFeature` when we want to avoid typescript
         * inlining these types during inference of `.d.ts` files.
         *
         * Mainly used where a Feature has dependencies:
         * ```ts
         * new Feature({
         *    id: 'someFeature',
         *    dependencies: [anotherFeature.asDependency],
         *    api: { ... },
         * })
         * ```
         */
        this.asDependency = this;
        this.environmentIml = new Set();
        this.setupHandlers = new patterns_1.SetMultiMap();
        this.contextHandlers = new Map();
        this.id = def.id;
        this.dependencies = def.dependencies || [];
        this.api = def.api || {};
        this.context = def.context || {};
        this.identifyApis();
    }
    /**
     * Call this to provide the environment specific implementation for a feature.
     *
     * @param env Environment id to implement the api for.
     * @param setupHandler Callback that receives:
     * - Own feature `Slot`s
     * - Dependencies APIs
     * - Context API that is specific to a runtime environment.
     *
     * @returns Implementation for the services defined for this feature on this environment
     */
    setup(env, setupHandler) {
        validateNoDuplicateEnvRegistration(env, this.id, this.environmentIml);
        this.setupHandlers.add(env.env, setupHandler);
        return this;
    }
    /**
     *
     *
     * @param config
     * @returns
     */
    use(config) {
        return [this.id, config];
    }
    setupContext(_env, environmentContext, contextHandler) {
        validateNoDuplicateContextRegistration(environmentContext, this.id, this.contextHandlers);
        this.contextHandlers.set(environmentContext, contextHandler);
        return this;
    }
    [symbols_1.CREATE_RUNTIME](runningEngine) {
        const { features, runOptions, referencedEnvs, entryEnvironment: { env: envName } } = runningEngine;
        const deps = {};
        const depsApis = {};
        const runningApi = {};
        const inputApi = {};
        const providedAPI = {};
        const environmentContext = {};
        const apiEntries = Object.entries(this.api);
        const feature = new RuntimeFeature(this, runningApi, deps);
        features.set(this, feature);
        for (const dep of this.dependencies) {
            deps[dep.id] = runningEngine.initFeature(dep);
            depsApis[dep.id] = deps[dep.id].api;
        }
        for (const [key, entity] of apiEntries) {
            const provided = entity[symbols_1.CREATE_RUNTIME](runningEngine, this.id, key);
            if (provided !== undefined) {
                inputApi[key] = provided;
            }
        }
        const settingUpFeature = {
            ...inputApi,
            id: this.id,
            run(fn) {
                feature.addRunHandler(fn, envName);
            },
            onDispose(fn) {
                feature.addOnDisposeHandler(fn, envName);
            },
            [symbols_1.RUN_OPTIONS]: runOptions,
            [symbols_1.ENGINE]: runningEngine,
            runningEnvironmentName: envName
        };
        for (const [key, contextHandler] of this.contextHandlers) {
            environmentContext[key] = { ...emptyDispose, ...contextHandler(depsApis) };
        }
        for (const envName of referencedEnvs) {
            const environmentSetupHandlers = this.setupHandlers.get(envName);
            if (!environmentSetupHandlers) {
                continue;
            }
            for (const setupHandler of environmentSetupHandlers) {
                const featureOutput = setupHandler(settingUpFeature, depsApis, environmentContext);
                if (!featureOutput) {
                    continue;
                }
                for (const [key, value] of Object.entries(featureOutput)) {
                    settingUpFeature[key] = value;
                }
                Object.assign(providedAPI, featureOutput);
            }
        }
        for (const [key, entity] of apiEntries) {
            const registered = entity[symbols_1.REGISTER_VALUE](runningEngine, providedAPI[key], inputApi[key], this.id, key);
            if (registered !== undefined) {
                runningApi[key] = registered;
            }
        }
        return feature;
    }
    identifyApis() {
        for (const [key, api] of Object.entries(this.api)) {
            const entityFn = api[symbols_1.IDENTIFY_API];
            if (entityFn) {
                entityFn.call(api, this.id, key);
            }
        }
    }
}
exports.Feature = Feature;
function validateNoDuplicateEnvRegistration(env, featureId, registered) {
    const hasCollision = (0, env_1.testEnvironmentCollision)(env, registered);
    if (hasCollision.length) {
        const collisions = hasCollision.join(', ');
        throw new Error(`Feature can only have single setup for each environment. ${featureId} Feature already implements: ${collisions}`);
    }
}
function validateNoDuplicateContextRegistration(environmentContext, featureId, contextHandlers) {
    const registeredContext = contextHandlers.get(environmentContext);
    if (registeredContext) {
        throw new Error(`Feature can only have single setupContext for each context id. ${featureId} Feature already implements: ${String(environmentContext)}`);
    }
}
//# sourceMappingURL=feature.js.map