"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isSetMultiMap = exports.SetMultiMap = void 0;
const common_1 = require("@wixc3/common");
/**
 * Maps keys to a set of values
 * @example
 * ```ts
 * const m = new SetMultiMap([['a',1],['a',2]])
 * m.add('a',3)
 * m.has('a',1) // => true
 * m.has('a',2) // => true
 * m.has('a',3) // => true
 * m.has('a',4) // => false
 * ```
 */
class SetMultiMap {
    constructor(entries) {
        this.map = new Map();
        (0, common_1.forEach)(entries, ([key, val]) => {
            this.add(key, val);
        });
    }
    get size() {
        return (0, common_1.chain)(this.map)
            .map(([_, { size }]) => size)
            .reduce((sum, size) => sum + size, 0).value;
    }
    get(key) {
        return this.map.get(key);
    }
    add(key, value) {
        const valueSet = this.map.get(key);
        if (valueSet) {
            valueSet.add(value);
        }
        else {
            this.map.set(key, new Set([value]));
        }
        return this;
    }
    clear() {
        this.map.clear();
    }
    delete(key, value) {
        const valueSet = this.map.get(key);
        if (valueSet) {
            const wasInSet = valueSet.delete(value);
            if (valueSet.size === 0) {
                this.map.delete(key);
            }
            return wasInSet;
        }
        return false;
    }
    deleteKey(key) {
        return this.map.delete(key);
    }
    has(key, value) {
        const valueSet = this.map.get(key);
        return valueSet ? valueSet.has(value) : false;
    }
    hasKey(key) {
        const existingSet = this.map.get(key);
        return !!existingSet && existingSet.size > 0;
    }
    [Symbol.iterator]() {
        return this.entries();
    }
    *entries() {
        const { map } = this;
        for (const [key, valueSet] of map.entries()) {
            for (const value of valueSet) {
                yield [key, value];
            }
        }
    }
    *values() {
        const { map } = this;
        for (const valueSet of map.values()) {
            yield* valueSet.values();
        }
    }
    keys() {
        return this.map.keys();
    }
}
exports.SetMultiMap = SetMultiMap;
function isSetMultiMap(x) {
    return x instanceof (SetMultiMap);
}
exports.isSetMultiMap = isSetMultiMap;
//# sourceMappingURL=set-multi-map.js.map