import { Signal } from './signal';
/**
 * A simple event emitter
 */
export class EventEmitter {
    constructor() {
        this.events = new Map();
        this.emitOnce = new Map();
        /**
         * Check if an event has subscribers
         * @returns true if event has subscribers
         */
        this.hasSubscribers = (event) => this.events.has(event);
        /**
         * Subscribe a handler for event
         */
        this.subscribe = (event, handler) => {
            const bucket = this.events.get(event);
            bucket ? bucket.add(handler) : this.events.set(event, new Signal([handler]));
        };
        /**
         * {@inheritDoc EventEmitter.subscribe}
         */
        this.on = this.subscribe;
        /**
         * Adds a handler that will be called at most once
         */
        this.once = (event, handler) => {
            this.off(event, handler);
            const bucket = this.emitOnce.get(event);
            bucket ? bucket.add(handler) : this.emitOnce.set(event, new Signal([handler]));
        };
        /**
         * Unsubscribe a handler from event
         */
        this.unsubscribe = (event, handler) => {
            let bucket = this.events.get(event);
            bucket === null || bucket === void 0 ? void 0 : bucket.delete(handler);
            (bucket === null || bucket === void 0 ? void 0 : bucket.size) === 0 && this.events.delete(event);
            bucket = this.emitOnce.get(event);
            bucket === null || bucket === void 0 ? void 0 : bucket.delete(handler);
            (bucket === null || bucket === void 0 ? void 0 : bucket.size) === 0 && this.events.delete(event);
        };
        /**
         * {@inheritDoc EventEmitter.unsubscribe}
         */
        this.off = this.unsubscribe;
        /**
         * Drop all subscriptions of a signal
         * @param event - signal id
         */
        this.delete = (event) => {
            this.events.delete(event);
            this.emitOnce.delete(event);
        };
        /**
         * Drop all subscriptions
         */
        this.clear = () => {
            this.events = new Map();
            this.emitOnce = new Map();
        };
        /**
         * {@inheritDoc Signal.notify}
         * @param event - eve
         * @param data - event data
         */
        this.notify = (event, data) => {
            var _a, _b;
            (_a = this.events.get(event)) === null || _a === void 0 ? void 0 : _a.notify(data);
            (_b = this.emitOnce.get(event)) === null || _b === void 0 ? void 0 : _b.notify(data);
            this.emitOnce.delete(event);
        };
        /**
         * {@inheritDoc EventEmitter.notify}
         */
        this.emit = this.notify;
    }
}
//# sourceMappingURL=event-emitter.js.map