"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.useScrollListPosition = void 0;
const react_1 = require("react");
const common_1 = require("../../common");
const use_position_1 = require("../../hooks/use-position");
const useScrollListPosition = ({ items, getId, getItemInfo, itemSize, isHorizontal, unmountItems = true, itemGap = 0, itemsInRow = 1, extraRenderSize = 0.5, averageItemSize, scrollWindowSize, itemsDimensions, scrollOffset = 0, maxScrollSize, scrollPosition, scrollListRef, }) => {
    const lastRenderedItem = (0, react_1.useRef)(0);
    const shouldMeasureOffset = typeof scrollOffset === 'number' ? common_1.unknownPosition : scrollOffset;
    const offsetFromParent = (0, use_position_1.usePositionInParent)(scrollListRef, shouldMeasureOffset);
    const usedOffset = (typeof scrollOffset === 'number' ? scrollOffset : isHorizontal ? offsetFromParent.x : offsetFromParent.y) || 0;
    //                Scroll List
    //            ┌─────────────────┐
    //            │                 │
    //            │                 │firstWantedPixel
    //            ├─────────────────┼────────────────────┐
    //            │                 │                    │
    //            │                 │                    │
    //            │                 │ ScrollWindow       │
    //       ┌────┴─────────────────┴─────────────┐      │
    //       │▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓│      │
    //       │    │                 │    ▲        │      │  shownItems
    //       │    │                 │    │        │      │◄─────────────────────
    //   ────┴───►│                 │    │ Offset │      │  to be passed to List
    //   Scrolled │                 │             │      │
    //       │    │                 │             │      │
    //       └────┼─────────────────┼─────────────┘      │
    //            │                 │                    │
    //            │                 │                    │
    //            │                 │ lastWantedPixel    │
    //   ┌► ──────┼─────────────────┼────────────────────┘
    //   │        │                 │ if accounted for
    //   │        │                 │
    //   │        │                 │ scroll position
    //   │        │                 │
    //   │        │                 │ and offset
    //   │        │                 │
    //   │        └─────────────────┘
    //   └─ Extra render size (0.5 for this example)
    // Last wanted pixel is either the position of scroll window with extra taking offset and scroll into account;
    // or the potential last pixel of the scroll list
    const lastWantedPixel = (0, react_1.useMemo)(() => {
        const scrollWindowSizeWithExtra = (1 + extraRenderSize) * scrollWindowSize;
        return Math.min(scrollWindowSizeWithExtra - usedOffset + scrollPosition, maxScrollSize);
    }, [extraRenderSize, scrollWindowSize, usedOffset, scrollPosition, maxScrollSize]);
    // First wanted pixel is counted backwards from last wanted pixel;
    // it can't be more than scrollWindow 2 times (as extra is between 0 and 1);
    const firstWantedPixel = (0, react_1.useMemo)(() => (unmountItems ? lastWantedPixel - 2 * scrollWindowSize : 0), [unmountItems, scrollWindowSize, lastWantedPixel]);
    return (0, react_1.useMemo)(() => {
        var _a, _b, _c;
        if (typeof itemSize === 'number') {
            const firstShownItemIndex = unmountItems ? Math.ceil(firstWantedPixel / itemSize) : 0;
            let lastShownItemIndex = Math.ceil(lastWantedPixel / itemSize);
            if (!unmountItems) {
                lastShownItemIndex = lastRenderedItem.current = Math.max(lastShownItemIndex, lastRenderedItem.current);
            }
            return {
                firstWantedPixel,
                firstShownItemIndex: firstShownItemIndex > 0 ? firstShownItemIndex : 0,
                lastShownItemIndex,
            };
        }
        let amountOfTakenPixels = -itemGap; // Initializing with -itemGap to compensate for adding gap for every item
        let firstTakenPixel = null;
        let firstShownItemIndex = 0;
        for (let rowIndex = 0; rowIndex < items.length; rowIndex += itemsInRow) {
            let maxRowSize = 0;
            for (let columnIndex = 0; columnIndex < itemsInRow && rowIndex + columnIndex < items.length; columnIndex++) {
                const item = items[rowIndex + columnIndex];
                const id = getId(item);
                const size = (_c = (itemSize === false
                    ? isHorizontal
                        ? (_a = itemsDimensions.current[id]) === null || _a === void 0 ? void 0 : _a.width
                        : (_b = itemsDimensions.current[id]) === null || _b === void 0 ? void 0 : _b.height
                    : itemSize === null || itemSize === void 0 ? void 0 : itemSize(getItemInfo(item)))) !== null && _c !== void 0 ? _c : averageItemSize;
                maxRowSize = Math.max(size, maxRowSize);
            }
            amountOfTakenPixels += maxRowSize + itemGap;
            if (unmountItems && amountOfTakenPixels > firstWantedPixel && firstTakenPixel === null) {
                firstTakenPixel = amountOfTakenPixels - maxRowSize;
                firstShownItemIndex = rowIndex;
            }
            if (amountOfTakenPixels > lastWantedPixel) {
                let lastShownItemIndex = rowIndex;
                if (!unmountItems) {
                    lastShownItemIndex = lastRenderedItem.current = Math.max(lastShownItemIndex, lastRenderedItem.current);
                }
                return {
                    firstWantedPixel: firstTakenPixel !== null && firstTakenPixel !== void 0 ? firstTakenPixel : 0,
                    firstShownItemIndex: firstShownItemIndex,
                    lastShownItemIndex,
                };
            }
        }
        const estimatedLastShownItemIndex = maxScrollSize / averageItemSize;
        return {
            firstWantedPixel: firstTakenPixel !== null && firstTakenPixel !== void 0 ? firstTakenPixel : 0,
            firstShownItemIndex: firstShownItemIndex,
            lastShownItemIndex: !isNaN(estimatedLastShownItemIndex)
                ? Math.max(items.length, estimatedLastShownItemIndex)
                : items.length,
        };
    }, [
        averageItemSize,
        firstWantedPixel,
        getId,
        getItemInfo,
        isHorizontal,
        itemGap,
        itemSize,
        items,
        itemsDimensions,
        itemsInRow,
        lastWantedPixel,
        maxScrollSize,
        unmountItems,
    ]);
};
exports.useScrollListPosition = useScrollListPosition;
//# sourceMappingURL=use-scroll-list-position.js.map